                 (* FUNCTEN.PAS   *)

unit Functen;

interface
                            
uses Enum_cons;

(* This file FUNCTEN.PAS contains procedure Func, which is called
   in procedure Update.
   This defines the function of the incidence matrix mat
   evaluated for every enumerated mat.
   The global variable g is used, which is the order of the matrix.
   The variables mat is declared as
     mat: Incid;
   where type Incid is declared as
     Incid: array[1..gmax,1..gmax] of Boolean;
  Some available variables are:
      mat      incidence matrix generated
      mut[h]   number of mutuals
      Sumfn[h,nf]
       sum over all (i,j), i >= j, of function Contr(nf) defined below
    (do not omit the index h; it's the number of the current last node)

   The function Tex gives a text of type
      Txt: string[50]
   which is printed on the screen for the identification of the
   evaluated function.
   Take good care that Tex agrees with Func; otherwise you will
   definitely make mistakes with output interpretation. *)

var funcerror: boolean;

function Numversion: byte;
(* Number of options for choosing a set of statistics. *)

function Numfunc(fversion: byte): Integer;
(* Number of statistics evaluated; should be not larger than Numfmax. *)

function Numsumfn: Byte;
(* Number of functions used in Sumfn *)

function Connected(var mb: Incid): Boolean;
(* Indicates whether mb is a (weakly) connected (di)graph; *)

procedure Func(fversion: byte; var fv: Funcvalue; var rf: rFuncvalue);
(* Statistics evaluated. *)
(* for i:= 1 to Numfunc, f[i] and rf[i] must be defined ! *)

procedure DatFunc(var dmat: Incid; fversion: byte; var fv: Funcvalue;
                  var rf: rFuncvalue);
(* Statistics evaluated; should be same as Func, but for a data matrix, i.e.,
   without using intermediate results for enumerated matrices as in Func. *)

function Tex(fversion: byte; nf: Byte; rounded: Boolean): shortstring;
(* Texts describing the statistics calculated in Func. *)
(* for i:= 1 to Numfunc, Tex[i] must be defined ! *)

procedure Contr(h: Byte);
(* This function gives the contribution to Triadfn of the pair
   (istar,jstar) with istar >= jstar ;
   it is assumed that it is called only when all mat[i,j] needed
   are indeed defined. This may need a check!
   If the desired function Func cannot be expressed as sum of such
   contributions, then do not use Sumfn[h] in Func, and
   make some easy assignment in Contr, e.g., Contr := 0;
   istar and jstar are given variables in the calling procedure *)

implementation

uses enum;

function Numversion: byte;
(* Number of options for choosing a set of statistics. *)
(* Allowed values of fversion are 1 to Numversion *)
begin
  if (f = g) then Numversion := 3 else Numversion := 2;
end;

function Numfunc(fversion: byte): Integer;
(* Number of statistics evaluated; should be not larger than Numfmax. *)
begin
  case fversion of
    1: Numfunc := 4;
    2: NumFunc := g;
    3: NumFunc := 4;
   end; (* case *)
end;

function Numsumfn: Byte;
(* Number of functions used in Sumfn *)
begin
  Numsumfn := 2;
end;

function Connected(var mb: Incid): Boolean;
(* Indicates whether mb is a connected (di)graph;
   if mb is not a graph, then it is symmetrized first. *)
(* should be used only for f = g *)
var i,j,k: byte; m,m2: Incid; pow: Integer; conn: Boolean;
begin
  m := mb;
  for i:=1 to g do m[i,i] := true;
  (* symmetrize : *)
  if (not graph) then for i:=1 to g do for j:=i-1 downto 1 do begin
    m[i,j] := (m[i,j] or m[j,i]);
    m[j,i] := m[i,j];
   end;
  pow := 1;
  m2 := m;
  repeat
    m := m2;
    pow := 2*pow;
    (* m2 := Boolean square of m : *)
    conn := true;
    for i:=g downto 1 do for j:=g downto 1 do begin
      k := 1;
      while ((not m2[i,j]) and (k <= g)) do begin
        if (m[i,k] and m[k,j]) then m2[i,j] := true;
        Inc(k);
       end; (* while *)
      conn := conn and m2[i,j];
     end; (* for i and j *)
   until ((pow >= g) or conn);
  Connected := conn;
end; (* Connected *)

procedure EigenVector(ma: Incid; var ev: Rvector);
(* Calculates eigenvector for largest eigenvalue of ma by power method;
   adds 1 to diagonal, and subtracts 1 again from the eigenvalue.
   ma is symmetrized first.
   The resulting vector ev is normed so that its length
   is the largest eigenvalue. *)
(* should be used only for f = g *)
const eps = 1e-6;
var i,j,k: byte; iev, iev2: Ivector; rev,rev2: Rvector;
    s,dif: real; maa: Incid;
begin
  maa := ma;
  (* symmetrize: *)
  for i:=1 to g do begin
    for j:=i-1 downto 1 do
      if (maa[i,j] or maa[j,i]) then
        begin maa[i,j] := true; maa[j,i] :=true; end;
    maa[i,i] := true;
   end; (* i *)
  for i:=1 to g do iev[i] := 1;
(* First some iterations in integer arithmetic; this increases speed. *)
  for k:=1 to kmax do begin
    for i:=1 to g do begin
      iev2[i] := 0;
      for j:=g downto 1 do if (maa[i,j]) then Inc(iev2[i],iev[j]);
     end; (* i *)
    iev := iev2;
   end; (* k *)
(* Now some iterations in real arithmetic *)
  for i:=1 to g do rev2[i] := iev[i]/1e6;
  repeat
    rev := rev2;
    s := 0.0; dif := 0.0;
    for i:=1 to g do begin
      rev2[i] := 0.0;
      for j:=g downto 1 do if (maa[i,j]) then
          rev2[i] := rev2[i]+rev[j];
      s := s + rev2[i];
     end; (* i *)
    for i:=1 to g do begin
      rev2[i] := rev2[i]/s;
      dif := dif + abs(rev[i]-rev2[i]);
     end; (* i *)
   until (dif < (g-1)*eps);
  s := sqrt(s);
  for i:=1 to g do ev[i] := (s-1)*rev2[i];
end; (* EigenVector *)

procedure TwoModeEigenVector(f,g: Byte; ma: Incid; var ev: Rvector);
(* Calculates eigenvector for largest eigenvalue of ma'*ma by power method;
   The resulting vector ev is normed to unit length. *)
const eps = 1e-6;
type gMat = array[1..gmax,1..gmax] of Byte;
var i,j,k,kmax: byte; iev, iev2: array[1..gmax] of integer;
    rev,rev2: array[1..gmax] of real;
    s,dif: real;
    maa: array[1..gmax,1..gmax] of byte;
begin
    for i:=1 to g do for j := 1 to g do maa[i,j] := 0;
    for i:=1 to g do for j := 1 to g do
      for k:=1 to f do if (ma[k,i] and ma[k,j]) then Inc(maa[i,j]);
  for i:=1 to g do iev[i] := 1;
(* First some iterations in integer arithmetic; this increases speed. *)
  kmax := trunc(ln(2.147e9)/ln(f*g));
  for k:=1 to kmax do begin
    for i:=1 to g do begin
      iev2[i] := 0;
      for j:=g downto 1 do iev2[i] := iev2[i] + iev[j] * maa[i,j];
     end; (* i *)
    iev := iev2;
   end; (* k *)
(* Now some iterations in real arithmetic *)
  for i:=1 to g do rev2[i] := iev[i]/1e6;
  repeat
    rev := rev2;
    s := 0.0; dif := 0.0;
    for i:=1 to g do begin
      rev2[i] := 0.0;
      for j:=g downto 1 do rev2[i] := rev2[i]+rev[j]*maa[i,j];
      s := s + rev2[i];
     end; (* i *)
    for i:=1 to g do begin
      rev2[i] := rev2[i]/s;
      dif := dif + abs(rev[i]-rev2[i]);
     end; (* i *)
   until (dif < (g-1)*eps);
  s := 0.0;
  for i:=1 to g do s := s + sqr(rev2[i]);
  s := 1.0/sqrt(s);
  for i:=1 to g do ev[i] := s*rev2[i];
end; (* TwoModeEigenVector *)

procedure Func(fversion: byte; var fv: Funcvalue; var rf: rFuncvalue);
(* Statistics evaluated. Should give same result as DatFunc below. *)
(* for i:= 1 to Numfunc, fv[i] and rf[i] must be defined ! *)
var iti,i1,i2,i3,fun1,fun2: Integer;
    i: Byte;
    ev: RVector;
(*    dfv: Funcvalue;
    drf: rFuncValue; *)
begin
 if (f = g) then begin
  case fversion of
    1: begin
        if (graph and not connect) then
            begin if Connected(mat) then fv[1] := 1 else fv[1] := 0; end
          else fv[1] := mut[h];
        fv[2] := Sumfn[h,1];
        fv[3] := Sumfn[h,2];
        for i:=1 to 3 do rf[i] := fv[i];
        if (Sumfn[h,1] + Sumfn[h,2] > 0) then begin
          rf[4] := 100*Sumfn[h,1]/(Sumfn[h,1] + Sumfn[h,2]);
          fv[4] := round(rf[4]);
         end else begin
          fv[4] := 0; rf[4] := 0.0;
         end;
       end; (* case 1 *)
    2: begin
         EigenVector(mat,ev);
         for i:=1 to g do begin
           fv[i] := round(ev[i]*100);
           rf[i] := ev[i]*100;
          end;
       end;  (* case 2 *)
     3: begin (* same as case 1 of f <> g *)
          fv[1] := ord(mat[1,1]);
          fv[2] := ord(mat[1,2]);
          fun1 := 0;
          for i:=g downto 1 do if (mat[1,i] and mat[2,i]) then inc(fun1);
          fv[3] := fun1;
          fun1 := 0;
          for i:=f downto 1 do if (mat[i,1] and mat[i,2]) then inc(fun1);
          fv[4] := fun1;
          for i:=1 to 3 do rf[i] := fv[i];
        end; (* case 3 *)
    end; (* case fversion *)
  end (* if f = g *) else begin
    case fversion of
     1: begin (* same as case 3 of f = g *)
          fv[1] := ord(mat[1,1]);
          fv[2] := ord(mat[1,2]);
          fun1 := 0;
          for i:=g downto 1 do if (mat[1,i] and mat[2,i]) then inc(fun1);
          fv[3] := fun1;
          fun1 := 0;
          for i:=f downto 1 do if (mat[i,1] and mat[i,2]) then inc(fun1);
          fv[4] := fun1;
          for i:=1 to 3 do rf[i] := fv[i];
        end; (* case 1 *)
   2: begin
         TwoModeEigenvector(f,g,mat,ev);
         for i:=1 to g do begin
           fv[i] := round(100*ev[i]);
           rf[i] := 100*ev[i];
          end (* for i *);
      end; (* case 2 *)
      end; (* case fversion *)
  end; (* if f <> g *)
(* For a check:
   Datfunc(mat,fversion,dfv,drf);
   for i:= numfunc(fversion) downto 1 do if (fv[i] <> dfv[i]) then begin
     writeln(outf,'Error: fv[',i,'] = ',fv[i],', dfv = ',dfv[i]);
     funcerror := true;
   end;         *)
end; (* Func *)

procedure DatFunc(var dmat: Incid; fversion: byte; var fv: Funcvalue;
                  var rf: rFuncvalue);
(* Statistics evaluated; should be same as Func, but for a data matrix, i.e.,
   without using intermediate results for enumerated matrices as in Func. *)
(* for i:= 1 to Numfunc, fv[i] and rf[i] must be defined ! *)
var i1,i2,i3,fun0,fun1,fun2: Integer;
    i: Byte;
    ev: RVector;
begin
 if (f = g) then begin
  case fversion of
    1: begin
         fun0 := 0; fun1 := 0; fun2 := 0;
         for i1 := 1 to g do
           for i2 := 1 to g do begin
             if ((i1 < i2) and dmat[i1,i2] and dmat[i2,i1]) then inc(fun0);
             for i3 := 1 to g do
               if (dmat[i1,i2] and dmat[i2,i3]) then begin
                 if dmat[i1,i3] then inc(fun1) else
                   if (i1 <> i3) then inc(fun2); end;
             end; (* for i1, i2 *)
         if (graph and not connect) then
            begin if Connected(dmat) then fv[1] := 1 else fv[1] := 0; end
          else fv[1] := fun0;
         fv[2] := fun1;
         fv[3] := fun2;
         for i :=1 to 3 do rf[i] := fv[i];
         if (fun1 + fun2 > 0) then
           rf[4] := 100.0*fun1/(fun1 + fun2)
          else rf[4] := 0.0;
         fv[4] := round(rf[4]);
       end; (* case 1 *)
    2: begin
         EigenVector(dmat,ev);
         for i:=1 to g do begin
           fv[i] := round(ev[i]*100);
           rf[i] := ev[i]*100;
          end;
       end;  (* case 2 *)
     3: begin    (* same as case 1 of f <> g *)
          fv[1] := ord(dmat[1,1]);
          fv[2] := ord(dmat[1,2]);
          fun1 := 0;
          for i:=g downto 1 do if (dmat[1,i] and dmat[2,i]) then inc(fun1);
          fv[3] := fun1;
          fun1 := 0;
          for i:=f downto 1 do if (dmat[i,1] and dmat[i,2]) then inc(fun1);
          fv[4] := fun1;
          for i:=1 to 3 do rf[i] := fv[i];
        end; (* case 3 *)
    end; (* case fversion *)
 end (* if f = g *) else begin
    case fversion of
     1: begin   (* same as case 3 of f = g *)
          fv[1] := ord(dmat[1,1]);
          fv[2] := ord(dmat[1,2]);
          fun1 := 0;
          for i:=g downto 1 do if (dmat[1,i] and dmat[2,i]) then inc(fun1);
          fv[3] := fun1;
          fun1 := 0;
          for i:=f downto 1 do if (dmat[i,1] and dmat[i,2]) then inc(fun1);
          fv[4] := fun1;
          for i:=1 to 3 do rf[i] := fv[i];
        end; (* case 1 *)
   2: begin
         TwoModeEigenvector(f,g,dmat,ev);
         for i:=1 to g do begin
           fv[i] := round(100*ev[i]);
           rf[i] := 100*ev[i];
          end (* for i *);
       end; (* case 2 *)
      end; (* case fversion *)
 end;
end; (* DatFunc *)

function Tex(fversion: byte; nf: Byte; rounded: Boolean): shortstring;
(* Texts describing the statistics calculated in Func. *)
(* for i:= 1 to Numfunc, Tex[i] must be defined ! *)
var Tex0, st: shortstring;
    roun: boolean;
begin
  roun := false;
  if (f = g) then begin
    case fversion of
      1: begin
            case nf of
            1:  if (graph and not connect) then
                  Tex0 := 'Connectivity (yes = 1, no = 0)'
                  else Tex0 := 'Number of mutual dyads';
            2:  Tex0 := '# of transitive relations in triads ';
            3:  Tex0 := '# of intransitive relations in triads ';
            4:  begin
                  Tex0 := '100 * #trans / (#trans + #intrans)';
                  roun := true;
                 end;
           end; (* case nf *)
         end; (* case 1 *)
      2: begin
           str(nf,st);
           Tex0 := '100 * eigenvector power index actor '+st;    
           roun := true;
         end; (* case 2 *)
     3: begin
          case nf of
          1..2: begin
               str(nf,st);
               Tex0 := 'Element (1,'+st+')';
              end; (* case 1 *)
          3: Tex0 := 'Matches of rows 1 and 2 ( = Sum_(j) Y_{1j} Y_{2j} )';
          4: Tex0 := 'Matches of columns 1 and 2 ( = Sum_(i) Y_{i1} Y_{i2} )';
          end; (* case nf *)
        end; (* case 3 *)
      end; (* case fversion *)
    end (* if f=g *) else begin
    case fversion of
     1: begin
          case nf of
          1..2: begin
               str(nf,st);
               Tex0 := 'Element (1,'+st+')';
              end; (* case 1 *)
          3: Tex0 := 'Matches of rows 1 and 2 ( = Sum_(j) Y_{1j} Y_{2j} )';
          4: Tex0 := 'Matches of columns 1 and 2 ( = Sum_(i) Y_{i1} Y_{i2} )';
          end; (* case *)
        end; (* case 1 *)
   2: begin
        str(nf:3,st);
        Tex0 := '100 * Eigenvector power index of column '+st+' (rounded to integer)';
        roun := true;
      end; (* case 2 *)
      end; (* case fversion *)
    end; (* if f <> g *)
  if (rounded and roun)
    then Tex := Tex0 + ' (rounded to integer)' else Tex := Tex0;
end; (* Tex *)

procedure Contr(h: Byte);
(* This function gives the contribution to Func of the pair
   (istar,jstar) with istar >= jstar ;
   it is assumed that it is called only when all mat[i,j] needed
   are indeed defined. This may need a check!
   If the desired function Func cannot be expressed as sum of such
   contributions, then do not use Sumfn[h] in Func, and
   make some easy assignment in Contr, e.g., Contr := 0;
   istar and jstar are given variables in the calling procedure *)
(* In the Contr between braces, Sumfn is a triad count,
   and Contr gives the contribution to the triad count
   of all triads istar,i,jstar with jstar < i < istar ;
   This triad count is the transitivity count *)
 var i,imax: Byte;
    tempi: array[1..2] of Integer;
begin
 if (f = g) then begin
  for i := nums downto 1 do tempi[i] := 0;
  if (istar < jstar) then
   (* nothing *) else
    if (istar <= jstar + 1) then
      (* nothing *) else begin
      imax := istar-1;
      for i:= jstar+1 to imax do begin
        if mat[istar,jstar] and mat[jstar,i] then
          if mat[istar,i] then tempi[1] := tempi[1] + 1
            else tempi[2] := tempi[2] + 1;
        if mat[istar,i] and mat[i,jstar] then
          if mat[istar,jstar] then tempi[1] := tempi[1] + 1
            else tempi[2] := tempi[2] + 1;
        if mat[jstar,istar] and mat[istar,i] then
          if mat[jstar,i] then tempi[1] := tempi[1] + 1
            else tempi[2] := tempi[2] + 1;
        if mat[jstar,i] and mat[i,istar] then
          if mat[jstar,istar] then tempi[1] := tempi[1] + 1
            else tempi[2] := tempi[2] + 1;
        if mat[i,istar] and mat[istar,jstar] then
          if mat[i,jstar] then tempi[1] := tempi[1] + 1
            else tempi[2] := tempi[2] + 1;
        if mat[i,jstar] and mat[jstar,istar] then
          if mat[i,istar] then tempi[1] := tempi[1] + 1
            else tempi[2] := tempi[2] + 1;
{        if mat[istar,jstar] and mat[jstar,i] then
          if mat[istar,i] then Inc(tempi[1]) else Inc(tempi[2]);
        if mat[istar,i] and mat[i,jstar] then
          if mat[istar,jstar] then Inc(tempi[1]) else Inc(tempi[2]);
        if mat[jstar,istar] and mat[istar,i] then
          if mat[jstar,i] then Inc(tempi[1]) else Inc(tempi[2]);
        if mat[jstar,i] and mat[i,istar] then
          if mat[jstar,istar] then Inc(tempi[1]) else Inc(tempi[2]);
        if mat[i,istar] and mat[istar,jstar] then
          if mat[i,jstar] then Inc(tempi[1]) else Inc(tempi[2]);
        if mat[i,jstar] and mat[jstar,istar] then
          if mat[i,istar] then Inc(tempi[1]) else Inc(tempi[2]);}
       end; (* i *)
      end; (* if *)
    sumfn[h,1] := sumfn[h,1] + tempi[1];
    sumfn[h,2] := sumfn[h,2] + tempi[2];
   end; (* if f = g *)
end; (* Contr *)

begin
 funcerror := false;
end.
