unit FunctSim;

interface

uses Simul, ZOMat;


function Numfunc(version,f,g:byte): Integer;
 (* Number of statistics evaluated. *)
function Numversion(f,g: byte): byte;
 (* number of possibilities for version *)
                          
(* The following procedures are for function evaluation.
   Function Func is called in procedure Update.
   This is the function of the random incidence matrix mat
   evaluated for every run.
   The function Tex gives a text which may be printed on the screen
   for the identification of the evaluated function.
   Take good care that Tex agrees with Func; otherwise you will
   definitely make mistakes with output interpretation. *)

procedure Func(var mat: Adjac; f,g,numl: Byte; fversion: Byte;
               graph, digraph, connect: Boolean; mutzero: Integer;
               rt: Marginr; var weit: TWeights;
               var f1: Funcvalue1; var f2: Funcvalue2;
               var rf1: rFuncvalue1);
 (* Evaluates statistics for matrix mat. *)

function Tex(nf: Integer; Texnumber: integer; fversion: Byte;
             graph, digraph, connect, round: Boolean): shortstring;
(* Gives a text describing the first and second calculated functions.
   May be useful for output of results. *)


implementation

uses Triads;

procedure EigenVector(f,g: Byte; ma: Adjac; var ev: Rvector);
(* Calculates eigenvector for largest eigenvalue of ma by power method;
   adds 1 to diagonal (does not affect eigenvector)
   ma is symmetrized first.
   The resulting vector ev is normed to unit length. *)
const eps = 1e-6;
var i,j,k,kmax: byte; iev, iev2: array[1..gmax] of integer;
    rev,rev2: array[1..gmax] of real;
    s,dif: real; maa: Adjac; ert: shortstring;
begin
(*  should only be used for f = g *)
  with maa do begin
    AllocateMat(g,g,ert);
    Defequal(ma,ert);
    (* symmetrize : *)
    for i:=1 to g do for j:=i-1 downto 1 do begin
      if (Arc(i,j) or Arc(j,i)) then begin
        Def(i,j,true); Def(j,i,true);
       end;
     end;
   end;
  for i:=1 to g do iev[i] := 1;
(* First some iterations in integer arithmetic; this increases speed. *)
  kmax := trunc(ln(2.147e9)/ln(g-1));
  for k:=1 to kmax do begin
    for i:=1 to g do begin
      iev2[i] := 0;
      for j:=g downto 1 do if (maa.Arc(i,j) or (i=j)) then Inc(iev2[i],iev[j]);
     end; (* i *)
    iev := iev2;
   end; (* k *)
(* Now some iterations in real arithmetic *)
  for i:=1 to g do rev2[i] := iev[i]/1e6;
  repeat
    rev := rev2;
    s := 0.0; dif := 0.0;
    for i:=1 to g do begin
      rev2[i] := 0.0;
      for j:=g downto 1 do if (maa.Arc(i,j) or (i=j)) then
          rev2[i] := rev2[i]+rev[j];
      s := s + rev2[i];
     end; (* i *)
    for i:=1 to g do begin
      rev2[i] := rev2[i]/s;
      dif := dif + abs(rev[i]-rev2[i]);
     end; (* i *)
   until (dif < (g-1)*eps);
  s :=0.0;
  for i:=1 to g do s := s + sqr(rev2[i]);
  s := 1.0/sqrt(s);
  for i:=1 to g do ev[i] := s*rev2[i];
  maa.Deallocatemat;
end; (* EigenVector *)

procedure TwoModeEigenVector(f,g: Byte; ma: Adjac; var ev: Rvector);
(* Calculates eigenvector for largest eigenvalue of ma'*ma by power method;
   The resulting vector ev is normed to unit length. *)
const eps = 1e-6;
type gMat = array[1..gmax,1..gmax] of Byte;
var i,j,k,kmax: byte; iev, iev2: array[1..gmax] of integer;
    rev,rev2: array[1..gmax] of real;
    s,dif: real; maa: gMat;
begin
  with ma do begin
    for i:=1 to g do for j := 1 to g do maa[i,j] := 0;
    for i:=1 to g do for j := 1 to g do
      for k:=1 to f do if (Arc(k,i) and Arc(k,j)) then Inc(maa[i,j]);
   end; (* with ma *)
  for i:=1 to g do iev[i] := 1;
(* First some iterations in integer arithmetic; this increases speed. *)
  kmax := trunc(ln(2.147e9)/ln(f*g));
  for k:=1 to kmax do begin
    for i:=1 to g do begin
      iev2[i] := 0;
      for j:=g downto 1 do iev2[i] := iev2[i] + iev[j] * maa[i,j];
     end; (* i *)
    iev := iev2;
   end; (* k *)
(* Now some iterations in real arithmetic *)
  for i:=1 to g do rev2[i] := iev[i]/1e6;
  repeat
    rev := rev2;
    s := 0.0; dif := 0.0;
    for i:=1 to g do begin
      rev2[i] := 0.0;
      for j:=g downto 1 do rev2[i] := rev2[i]+rev[j]*maa[i,j];
      s := s + rev2[i];
     end; (* i *)
    for i:=1 to g do begin
      rev2[i] := rev2[i]/s;
      dif := dif + abs(rev[i]-rev2[i]);
     end; (* i *)
   until (dif < (g-1)*eps);
  s := 0.0;
  for i:=1 to g do s := s + sqr(rev2[i]);
  s := 1.0/sqrt(s);
  for i:=1 to g do ev[i] := s*rev2[i];
end; (* TwoModeEigenVector *)


function Numversion(f,g: byte): byte;
(* Number of options for choosing a set of statistics. *)
(* Permitted values of fversion are 1 to Numversion *)
begin
  if (f=g) then Numversion := 4 else Numversion := 3;
end;

function Numfunc(version,f,g:byte): Integer;
(* Number of statistics evaluated. *)
begin
  case version of
    1: Numfunc := 3;
    2: Numfunc := g;
    3: Numfunc := g;
    4: Numfunc := 3;
    else Numfunc := 0;
   end; (* case *)
end;

procedure Func(var mat: Adjac; f,g,numl: Byte; fversion: Byte;
               graph, digraph, connect: Boolean; mutzero: Integer;
               rt: Marginr; var weit: TWeights;
               var f1: Funcvalue1; var f2: Funcvalue2;
               var rf1: rFuncvalue1);
(* Statistics evaluated. *)
(* for h:= 1 to numf = Numfunc, f1[h] and rf1[h] must be defined;
   for h:=1 to  numl2, f2[h] must be defined ! *)
(* In Simulate, the number of mutuals is also available as the variable mut.
   If only the number of mutuals were required and not the proportion
   of transitivity, then it would be more efficient to directly borrow
   the variable mut. *)
(* The variable rt = vector of row totals is used as an argument for Func
   for possible use in function calculation. *)

var i, j, k, maxi: Integer; temp1, temp2, temp3: integer; temp11: real;
  ev: RVector;
  tc: TriadCountType;
begin
(* In view of procedures Update and Giveresult, f1[h] must have
   non-negative integer results, not larger than m2 *)
 if (f = g) then begin
  case fversion of
    1: begin
         temp1 := 0;
         temp2 := 0;
         temp3 := 0;
         with mat do
         for i:= f downto 2 do
           for j:= i-1 downto 1 do begin
             if (Arc(i,j) and Arc(j,i)) then Inc(temp1);
             for k:= j-1 downto 1 do begin
               if Arc(i,j) and Arc(j,k) then
                 if Arc(i,k) then Inc(temp2) else Inc(temp3);
               if Arc(i,k) and Arc(k,j) then
                 if Arc(i,j) then Inc(temp2) else Inc(temp3);
               if Arc(j,i) and Arc(i,k) then
                 if Arc(j,k) then Inc(temp2) else Inc(temp3);
               if Arc(j,k) and Arc(k,i) then
                 if Arc(j,i) then Inc(temp2) else Inc(temp3);
               if Arc(k,i) and Arc(i,j) then
                 if Arc(k,j) then Inc(temp2) else Inc(temp3);
                 if Arc(k,j) and Arc(j,i) then
                 if Arc(k,i) then Inc(temp2) else Inc(temp3);
              end; (* k *)
            end; (* j *)
         if (graph and (not connect)) then begin
           if Connected(f,g,graph,mat) then f1[1] := 1 else f1[1] := 0;
           rf1[1] := f1[1];
          end else begin
           f1[1] := temp1 - mutzero;
           rf1[1] := temp1 - mutzero;
          end;
         if (temp2 + temp3 > 0) then begin
            f1[2] := round(100.0*temp2/(temp2+temp3));
            rf1[2] := temp2/(temp2+temp3);
           end else begin
            f1[2] := 0;
            rf1[2] := 0.0;
           end;
          with mat do if Arc(newone,newtwo) then begin
           f1[3] := 1; rf1[3] := 1.0; end else begin
           f1[3] := 0; rf1[3] := 0.0; end;
        end; (* case 1 *)
    2: begin
         if (g <= numfplus) then maxi := g else maxi := numfplus;
         EigenVector(f,g,mat,ev);
         for i:=1 to maxi do begin
           f1[i] := round(100*ev[i]);
           rf1[i] := 100*ev[i];
          end (* for i *);
       end; (* case 2 *)
    3: begin
         if (g <= numfplus) then maxi := g else maxi := numfplus;
         EigenVector(f,g,mat,ev);
         for i:=1 to maxi do begin
           f1[i] := round(1000*ev[i]);
           rf1[i] := 1000*ev[i];
          end (* for i *);
       end; (* case 3 *)
    4: begin (* identical to case 1 of f <> g *)
         with mat do begin
          if Arc(newone,newone) then begin
           f1[1] := 1; rf1[1] := 1.0; end else begin
           f1[1] := 0; rf1[1] := 0.0; end;
          if Arc(newone,newtwo) then begin
           f1[2] := 1; rf1[2] := 1.0; end else begin
           f1[2] := 0; rf1[2] := 0.0; end;
          temp1 := 0;
          for i:=g downto 1 do if (Arc(newone,i) and Arc(newtwo,i))
            then Inc(temp1);
          f1[3] := temp1; rf1[3] := round(temp1);
         end; (* with mat *)
       end; (* case 4 *)
   end; (* case fversion *)
  if (numl > 0) then begin TriadCount(mat,g,tc);
    for i:=1 to numl do begin
      temp11 := 0.0;
      for j := 1 to 16 do temp11 := temp11 + weit[i,j]*tc[j];
      f2[i] := temp11;
      if (temp11 >= weit[i,0]) then f2[numl+i] := 1.0 else f2[numl+i] := 0.0;
     end; (* i *)
   end; (* if numl *)
 end (* if f=g *) else begin
  case fversion of
    1: begin (* identical to case 4 of f = g *)
         with mat do begin
          if Arc(newone,newone) then begin
           f1[1] := 1; rf1[1] := 1.0; end else begin
           f1[1] := 0; rf1[1] := 0.0; end;
          if Arc(newone,newtwo) then begin
           f1[2] := 1; rf1[2] := 1.0; end else begin
           f1[2] := 0; rf1[2] := 0.0; end;
          temp1 := 0;
          for i:=g downto 1 do if (Arc(newone,i) and Arc(newtwo,i))
            then Inc(temp1);
          f1[3] := temp1; rf1[3] := round(temp1);
         end; (* with mat *)
       end; (* case 1 *)
    2: begin
         TwoModeEigenvector(f,g,mat,ev);
         if (g <= numfplus) then maxi := g else maxi := numfplus;
         for i:=1 to maxi do begin
           f1[i] := round(100*ev[i]);
           rf1[i] := 100*ev[i];
          end (* for i *);
          end;
    3: begin
         TwoModeEigenvector(f,g,mat,ev);
         if (g <= numfplus) then maxi := g else maxi := numfplus;
         for i:=1 to maxi do begin
           f1[i] := round(1000*ev[i]);
           rf1[i] := 1000*ev[i];
          end (* for i *);
          end;
   end; (* case *)
 end; (* if not graph or digraph *)
end; (* Func *)

function Tex(nf, Texnumber: Integer; fversion: Byte;
             graph, digraph, connect, round: Boolean): shortstring;
(* Description of statistics evaluated.
   Don't forget to define OrderDependent in the correct way.
   The order-dependent statistics should always be the last
   of the nf cases. *)
var Tex0, ttt: shortstring;
    roun: boolean;
begin
 OrderDependent := false;
 roun := false;
 if (f = g) then begin
  case fversion of
  1: begin
     case nf of
      1:   if (graph and not connect) then
                  Tex0 := 'Connectivity (yes = 1, no = 0)'
             else begin
               str(Texnumber,ttt);
               if (Texnumber = 0) then
               Tex0 := 'Number of mutual dyads' else
               Tex0 := 'Number of mutual dyads minus '+ttt;
              end;
      2:  begin
            Tex0 := '100 * #trans / (#trans + #intrans)';
            roun := true;
           end;
      3:  begin
            Tex0 := 'Arc indicator from 1 to 2';
            OrderDependent := true;
           end;
        end; (* case nf *)
     end; (* case 1 *)
  2: begin
      str(nf:3,ttt);
      Tex0 := '100 * Eigenvector power index of actor '+ttt;
      roun := true;
     end; (* case 2 *)
  3: begin
      str(nf:3,ttt);
      Tex0 := '1000 * Eigenvector power index of actor '+ttt;
      roun := true;
     end; (* case 3 *)
   4: begin (* identical to case 1 of f <> g *)
        case nf of
        1..2: begin
             str(nf,ttt);
             Tex0 := 'Element (1,'+ttt+')';
            end; (* case 1 *)
        3: Tex0 := 'Matches of rows 1 and 2 ( = Sum_(j) Y_{1j} Y_{2j} )';
        end; (* case *)
       OrderDependent := true;
      end; (* case 4 *)
  end; (* case fversion *)
 end (* if f = g *) else begin
  case fversion of
   1: begin (* identical to case 4 of f = g *)
        case nf of
        1..2: begin
             str(nf,ttt);
             Tex0 := 'Element (1,'+ttt+')';
            end; (* case 1 *)
        3: Tex0 := 'Matches of rows 1 and 2 ( = Sum_(j) Y_{1j} Y_{2j} )';
        end; (* case *)
       OrderDependent := true;
      end; (* case 1 *)
   2: begin
        str(nf:3,ttt);
        Tex0 := '100 * Eigenvector power index of column '+ttt;
        roun := true;
      end;
   3: begin
        str(nf:3,ttt);
        Tex0 := '1000 * Eigenvector power index of column '+ttt;
        roun := true;
      end;
  end; (* case *)
 end; (* not graph or digraph *)
 if (roun and round) then Tex0 := Tex0 + ' (rounded to integer)';
 if (OrderDependent and ChangedOrder) then
   Tex := Tex0 + ' (for original vertex order)' else Tex := Tex0;
end; (* Tex *)

end.
