                 (* NUMBERS.PAS         *)

unit Numbers;
(* These functions assume
 the constant fgmax
 the type Margin = array[1..fgmax] of Byte  *)

interface

uses Enum_cons;

function Bender(f,g: Byte; a: Marginr; b: Marginc; zero: Incid): Real;
(* The logarithm of the approximate number of fxg 0-1 matrices
   with row sums a, column sums b, structural zeros in zero.
   See Bender, Discrete Mathematics 10 (1974), 217-223 *)

function Bipart
  (a,b: Margin; g: Byte; var ErrInd: byte): Real;
(* This function calculates the number of bipartite graphs with g points
   in both subsets and a, b as vectors of degrees.
   Otherwise said, this is the number of gxg matrices with entries 0 or 1,
   with a and b as marginal sum vectors.
   See Sukhatme (Phil. Trans. Roy. Soc. London, A237,1938,375).
   An earlier Turbo Pascal version of this procedure checked dynamic memory
   using the MemAvail function, which is not available in Delphi.
   Therefore this function can now lead to memory errors.
   Therefore it is best, when calling this function or the next one,
   to protect it by putting it inside a try .. except statement.  *)

function Ndigraph
  (a,b: Margin; g: Byte; var ErrInd: byte): Real;
(* This function calculates the number of digraphs with vectors of
   out-degrees and in-degrees a, resp. b  .
   See Katz and Powell, Proc. Am. Math. Soc. 5, 1954, 621-626
                        Ann. Math. Stat. 28, 1957, 442-448.
   If ErrInd is positive on exit, then the free memory available
   was insufficient and the function value returned is meaningless. *)

implementation

(* the following variables are used for checking whether
    enough memory is available at the heap : *)
var needed, amount3, amount4: cardinal;
  HeapStat: THeapStatus;
  first: boolean;

procedure display1(i: longint);
begin
  (* display i on the screen; progress *)
end;

procedure display2(x: real);
var i: longint;
begin
  (* display i on the screen; max of display1 *)
end;
                    
procedure display3(x: real);
var i: longint;
begin
  (* display i on the screen; max of display1 *)
end;
function Bipart
  (a,b: Margin; g: Byte; var ErrInd: byte): Real;
(* This function calculates the number of bipartite graphs with g points
   in both subsets and a, b as vectors of degrees.
   Otherwise said, this is the number of gxg matrices with entries 0 or 1,
   with a and b as marginal sum vectors.
   See Sukhatme (Phil. Trans. Roy. Soc. London, A237,1938,375).
   If ErrInd is positive on exit, then free memory was insufficient for
   the calculations and the function value is meaningless.
   You may consult page 345 etc. of the Turbo Pascal 4 manual for the way
   in which the program checks free memory availability.
   The pointer heap consists of two subheaps:
   the first, "old", heap contains partitions left after dealing with row i-1;
   the second, "new", heap contains partitions created by dealing with row i.
   At the start of the program, the old heap consists of only partition b,
   and the new heap is empty.
   At the end of the program, the old heap consists of the zero partition. *)

type
  partition = array[0..gmax] of Byte;
   (* elements 1..g contain the partition;
      element 0 contains an upper bound to the number of nonnull elements.
      See procedure Orderpart *)
  subset    = array[1..gmax] of Boolean;
  partitionpointer = ^partitionrecord;
  partitionrecord = record
                      amount: Real;
                      part  : Partition;
                      next  : PartitionPointer;
                    end;
var
  firstold, nextold, lastold, lastnew: PartitionPointer;
  i, ai : Byte;
  aa,bb: Partition;
  FreeCount, FreeSpace: word;

procedure firstones(k: Byte; var ones: Subset);
var i: Byte;
begin
  for i:= k downto 1 do ones[i] := true;
  for i:= k+1 to gmax do ones[i] := false;
end;

procedure nextones(k: Byte; var ones: Subset; var last: Boolean);
(* the first pattern (1 0) in ones[1] through ones[k] is reversed into (0 1);
   the 1's before this pattern are shifted to the first places *)
var i, n0, n1: Byte;
begin
  n0 := 0;
  repeat
    n0 := succ(n0)
   until ones[n0];
  n1 := n0;
  repeat
    n1 := succ(n1)
   until not ones[n1];
  if n1 > k then
    begin last := true; Exit; end
   else last := false;
  ones[n1] := true;
  ones[pred(n1)] := false;
  n1 := n1 - 2;
  n0 := n1 - (n0 - 1);
  for i := n0 downto 1 do ones[i] := true;
  for i := n0 + 1 to n1 do ones[i] := false;
end; (* nextones *)

procedure minus(ones: Subset; part: Partition;
                  var c: Partition; var ErrInd1: byte);
var j: Integer;
begin
  if (part[0] > g) then begin
    ErrInd1 := 1;
    Exit;
   end else ErrInd1 := 0;
  for j := part[0] downto 1 do
    begin
      if ones[j] then
         c[j] := pred(part[j]) else
         c[j] := part[j];
      end; (* j *)
  c[0] := part[0];
end; (* minus *)

procedure Checknewps(y: Real; c: Partition; var neednew: Boolean;
                      var ErIn: byte);
(* checks if partition c is already in the new heap *)
var different: Boolean;
    i,k : Byte;
    thisnew: PartitionPointer;
begin
  ErIn := 0;
  k := c[0];
  thisnew := lastold^.next;
  while thisnew <> nil do
    with thisnew^ do begin
      (* check if c = thisnew^.part *)
      if k = part[0] then begin
        if k > 0 then begin
          i := k;
          repeat
            different := part[i] <> c[i];
            i := pred(i);
          until different or (i < 1);
         end else (* k = part[0] = 0 *) different := false;
        end else (* k <> part[0] *) different := true;
        if different then thisnew := next
          else begin
            amount := amount + y;
            neednew := false;
            exit;
           end;
        end; (* with thisnew *)
     neednew := true;
end; (* Checknewps *)

procedure Orderpart(var p: partition);
var i,j,k: Byte;
    q: partition;
begin
  for i := g downto 1 do q[i] := 0;
  for i := p[0] downto 1 do
    if p[i] > 0 then q[p[i]] := succ(q[p[i]]);
  k := 0;
  for i := g downto 1 do begin
    j := q[i];
    while j > 0 do begin
      k := succ(k);
      p[k] := i;
      j := pred(j);
     end;
    end; (* for i *)
  for i := k+1 to g do p[i] := 0;
  p[0] := k;
end; (* orderpart *)


procedure Inherit(firstold: PartitionPointer; k: Byte; var Erin:byte);
(* computes all partitions which are inheritors of partition firstold *)
var neednew,last: Boolean;
    y: Real;
    p0, Erin1 : Byte;
    ones: Subset;
    c: Partition;
    newnew: PartitionPointer;
    FreeSpace, FreeCount: longint;
begin
  Erin := 0;
  with firstold^ do
   if k <= part[0] then
    begin
      y := amount;
      p0 := part[0];
      Firstones(k,ones);
      Repeat
        minus(ones,part,c,Erin1);
        if (Erin1 > 0) then begin
          ErIn := 1;
          Exit;
         end;
        orderpart(c);
        checknewps(y,c,neednew,ErIn1);
        if (ErIn1 > 0) then begin
          exit;
          ErIn := 2;
          Exit;
         end;
        if neednew then begin
          HeapStat := GetHeapStatus;
          if first then
            amount3 := HeapStat.TotalFree
           else
(* The following criterion has been determined by trial and error.
   It may not always work.
   If a better programmer has a better alternative,
   please tell me about it! T.S. *)
            if ((HeapStat.Freebig ) < (needed)) then begin
             ErIn := 2;
             exit;
            end;
(* Turbo 4.0 version:      FreeSpace := Ofs(FreePtr^) - Ofs(HeapPtr^);
          FreeSpace := MemAvail;
          if (FreeSpace <= 1) then begin
            ErIn := 2;
            Exit;
           end;  *)
          New(newnew);
          if first then begin
            HeapStat := GetHeapStatus;
            amount4 := HeapStat.TotalFree;
            needed := (g+1)*(amount3 - amount4);
            first := false;
           end;
          lastnew^.next := newnew;
          newnew^.amount := y;
          newnew^.part := c;
          newnew^.next := nil;
          lastnew := newnew;
         end; (* if neednew *)
        Nextones(p0,ones,last);
       until last;
    end; (* with firstold if k <= part[0] *)
    (* if k > part[0] then there are no inheritors *)
end; (* Inherit *)

begin (* Bipart *)
  BiPart := 0.0;
  ErrInd := 0;
  bb[0] := g;
  for i := 1 to g do
    begin aa[i] := a[i]; bb[i] := b[i]; end;
  orderpart(bb);
  New(Firstold);
  with firstold^ do begin
    amount := 1.0;
    part := bb;
    next := nil;
  end;
  lastold := firstold;
  lastnew := firstold;
  for i:= 1 to g do begin
    display1(i);  (* indicate how far the program has come *)
    ai := aa[i];
    if ai > 0 then begin
      repeat
        inherit(firstold,ai,ErrInd);
        if (ErrInd > 0) then exit;
        nextold := firstold^.next;
(* In the Turbo 4.0 version:
        FreeCount := (8192 - Ofs(FreePtr^) div 8) mod 8192;
        FreeSpace := Ofs(FreePtr^) - Ofs(HeapPtr^);
        if (FreeCount > 8190) then ErrInd := 3;
        if (FreeSpace <= FreeMin) then ErrInd := 4;
*)
        dispose(firstold);
        firstold := nextold;
       until firstold = lastold^.next;  (* now the old heap is empty *)
      lastold := lastnew; (* now the old heap is the former new heap *)
     end; (* ai > 0 *)
   end; (* i *)
  if firstold <> nil then bipart := firstold^.amount else bipart := 0.0;
(* In the Turbo 4.0 version:
    Release(HeapOrg);
  Alternative for the use of Release:
  FreeCount := (8192 - Ofs(FreePtr^) div 8) mod 8192;
  FreeSpace := Ofs(FreePtr^) - Ofs(HeapPtr^);
  if (FreeCount > 8190) then ErrInd := 5;
  if (FreeSpace <= FreeMin) then ErrInd := 6;
  if (ErrInd > 0) then Exit;
  while firstold <> nil do begin
    nextold:= firstold^.next;
    dispose(firstold);
    firstold:=nextold;
   end; *)   
end; (* bipart *)

function Ndigraph
  (a,b: Margin; g: Byte; var ErrInd: byte): Real;
(* This function calculates the number of digraphs with vectors of
   out-degrees and in-degrees a, resp. b  .
   See Katz and Powell, Proc. Am. Math. Soc. 5, 1954, 621-626
                        Ann. Math. Stat. 28, 1957, 442-448.
   If ErrInd is positive on exit, then the free memory available
   was insufficient and the function value returned is meaningless. *)

var powers, minab, aa, bb: Margin;
    i, erind: Byte;
    x,count,tcount: Real;
    last, parity: Boolean;
    
procedure Subtract(powers: Margin; var aa,bb: Margin; var parity: Boolean);
var i,j,pi: Byte;
begin
  j := 0;
  for i := g downto 1 do
    begin
      pi := powers[i];
      j := j + pi;
      aa[i] := a[i] - pi;
      bb[i] := b[i] - pi;
     end;
   parity := not odd(j);
end;

Procedure Nextpws(var powers: Margin; var last: Boolean);
var i,j: Byte;
begin
  i := 1;
  while ( (i <= g) and (powers[i] >= minab[i])) do i := succ(i);
  if i > g then last := true else begin
    last := false;
    powers[i] := succ(powers[i]);
    for j := i-1 downto 1 do powers[j] := 0;
   end;  (* if i<=g *)
 end; (* Nextpws *)

begin  (* Ndigraph *)
  ErrInd := 0;
  for i:= g downto 1 do
    begin
      powers[i] := 0;
      if a[i] < b[i] then minab[i] := a[i] else minab[i] := b[i];
    end;
  x := 0.0;
  tcount := 1.0;
  for i := 1 to g do tcount := tcount * (minab[i] + 1);
  count := 0.0;
  display2(tcount); (* maximum number of calculations *)
  repeat
    Subtract(powers,aa,bb,parity);
    if parity then x := x + Bipart(aa,bb,g,erind)
      else x := x - Bipart(aa,bb,g,erind);
    if (erind > 0) then begin
      ErrInd := erind;
      Ndigraph := 0.0;
      Exit;
     end; (* if *)
    Nextpws(powers,last);   
    count := count + 1.0;
    display3(count);
   until last;
  Ndigraph := x;
//  closefile(af);
end;  (* Ndigraph  *)

function Bender(f,g: Byte; a: Marginr; b: Marginc; zero: Incid): Real;
(* fmax and gmax are constants;
   Marginr = array[1..fmax] of Byte; Marginc = array[1..gmax] of Byte;
   incid =array[1..fmax,1..gmax] of Boolean *)
(* The logarithm of the approximate number of fxg 0-1 matrices
   with row sums a, column sums b, structural zeros in zero.
   See Bender, Discrete Mathematics 10 (1974), 217-223 *)
var i1, i2, i3, i4, i, j: Integer;
    lf1, lf2: Real;

function lf(a: Byte): Real;
var i: Byte;
    t: Real;
begin
  t := 0;
  for i := a downto 1 do t := t + ln(i);
  lf := t;
end; (* lf *)

begin
  i1 := 0; i2 := 0; i3:= 0; i4 := 0;
  lf1 := 0; lf2 := 0;
  for j:= g downto 1 do begin
    i3 := i3 + sqr(b[j]);
    lf2 := lf2 + lf(b[j]);
   end; (* j *)
  for i := f downto 1 do begin
    i1 := i1 + a[i];
    i2 := i2 + sqr(a[i]);
    lf1 := lf1 + lf(a[i]);
    for j := g downto 1 do
      if (zero[i,j]) then i4 := i4 + a[i]*b[j];
   end; (* i *)
  if (i1 > 0) then
    Bender := lf(i1) - lf1 - lf2 - (i2 - i1)*(i3 - i1)/(2*sqr(i1)) - i4/i1
   else Bender := 0;
end; (* Bender *)

begin
 first := true;
end.
