                             (* SIMUL.PAS *)


(* Simulation procedures
   Programmed by Tom Snijders, University of Groningen *)

unit Simul;

interface

uses ZOMat, Triads, Rand3;

const
  m1 = -2;          (* m1 and m2 are bounds of arrays used for    *)
  m2 = 100;         (* function evaluation                        *)
  numfmax = 4;
  (* numfmax is maximum number of functions evaluated *)
  numfplus = 200;    (* Related to fmax and gmax, defined in ZOMAT;
                       at this moment, let numfplus = gmax .
                       see function Numfunc. *)
  numlmax = 18;     (* maximum of triad count functions *)
  numl2max = 36;    (* must be 2*numlmax *)
type

(* for simulation : *)
  Marginc = array[1..gmax] of Byte;
  Marginr = array[1..fmax] of Byte;    
(* for calculating function values : *)
  Funcvalue1 = array[1..numfplus] of Integer;
  RFuncvalue1 = array[1..numfplus] of real;
  Funcvalue2 = array[1..numl2max] of Real;
  TWeights = array[1..numlmax,0..16] of real;
  RVector = array[1..gmax] of real;
(* for bookkeeping of results: *)
  Res  = array[1..numfmax,m1..m2] of real;
  ResB = array[1..numfmax] of Boolean; (* idem *)
  ResI = array[1..numfmax] of Integer; (* idem *)
  ResL = array[1..numl2max] of real; (* idem *)
  ResL2 = array[1..numlmax,1..numlmax] of real;
  Res1 = array[m1..m2] of Real; (* used in Giveresult etc. *)

(* Global variables for simulations : *)
var
(* for input/output : *)
  f, g, numl, mode: Byte;
  nrun, nsuc: Integer;
  ntry: Int64;
  rt: Marginr; ct: Marginc; (* prescribed row margins *)
  mutzero,presmuts: Integer;
         (* origin for, and prescribed numbers of mutuals *)
  ChangedOrder, OrderDependent: Boolean;
         (* ChangeOrder indicates whether vertices were reordered,
            OrderDependent whether the statistics are order-dependent. *)
  newone, newtwo: Integer;
    (* vertex numbers to which vertices 1 and 2 are reordered *)       
  strzer: Adjac; (* prescribed matrix of structural zero indicators *)
  s, s2, c: Res;             (* sums and sums of squares of results *)
  sl, s2l, cl: ResL;         (* idem *)
  ssl: ResL2;                (* sums of results *)
  sp, sp2: real;             (* sum and sum of squares of npra *)
  meanp,varp: real;          (* final results *)
  mean,sd,cummean,cumsd: Res;(* final results *)
  meanl,sdl: ResL;           (* final results *)
  covl: ResL2;               (* final results *)
  large, small: ResB;        (* final results *)
  weit: TWeights;            (* weights for lin comb of triad counts *)
  errtext, warntext: shortstring;

  printall, useresfile, printfileconn: Boolean;
  outf, pfile, resfile: Textfile;
               (* resfile is file for intermediate results *)
  pname: shortstring; (* printfile name *)
  numf, numl2: byte; (* numbers of evaluated functions *)
  fversion: byte;    (* which functions evaluated in Func *)
  Normfac: real;  (* normalization factor of probabilities *)
  halted, haltit: boolean;

procedure unpackmode(var full,digr,graph,gnrl,
                         mutfix,rorder,connect,MolReed: boolean);

function sumr(var a: Marginc; g: Byte): integer ;
function sumc(var a: Marginr; f: Byte): integer ;

function Connected(f,g: Byte; graph: boolean; mat: Adjac): Boolean;
(* Indicates whether mat is a connected (di)graph;
   if mat is not a graph, then it is symmetrized first. *)
(* should only be used for f=g *)

procedure Displays(nt1,nt2:integer);

procedure Calcresult(nrun: integer; numf, numl, numl2: Byte;
(* input arrays and reals : *)
        var sp,sp2: real; var s,s2,c: Res;
        var sl,s2l,cl: ResL; var ssl: ResL2;
(* output arrays and reals : *)
        var meanp,varp: real; var mean,sd,cummean,cumsd: Res;
        var meanl,sdl: ResL; var covl: ResL2);
(* This procedure must follow Simulate, when the desired number of runs
   has been executed. *)

procedure Simulate(first: Boolean;
(* indicates whether this is the first of a simulation job,
   or continues a simulation job that has already started ; *)
           f,g: Byte;
(* dimensions of the matrix *)
           mode: Byte;
(* mode of matrix:
   last digit is called basemode (1-7);
    mode = 1: matrix without structural zeroes;
           2: square matrix with structurally zero diagonal (digraph);
           3: symmetric matrix with structurally zero diagonal (graph);
           4: arbitrary pattern, structural zeroes in matrix strzer;
           5: as 2, with number of mutual dyads = symmetric (1,1) pairs
                    fixed at the value presmuts;
           6: as 2, restriction to connected digraphs;
           7: as 3, restriction to connected graphs;
           12,13,15,16,17: as 2,3,5,6,7 with rows and columns
              to be reordered according to decreasing values
              of |g - 1 - 2*rt[i]|
           same + 100: Molloy-Reed algorithm (only if basemode 1,2,3)   *)
           var strzer: Adjac; presmuts: Integer;
           var rt: Marginr; var ct: Marginc;
(* see above; for basemode = 3 or 7
   the value of the array ct has no input meaning
   and a graph with row and column sums as in rt will be simulated. *)
(* prescribed vectors of row sums and column sums *)
           var weit: TWeights;
(* weights for linear combinations of triad counts *)
           numl: Byte;
(* number of functions of triad counts evaluated *)
           var nrun: integer;
(* desired number of runs *)
           var errtext, warntext: shortstring;
(* if, after exiting from Simulate, errtext = '', then the procedure has
           proceeded correctly;if not, errtext indicates what went wrong;
           warntext can contain a warning *)
           var mutzero: integer;
(* origin for number of mutuals:
   distribution of first reported function in output Res arrays
   is distribution of mut minus mutzero;
   this transformation does not affect the prescribed value presmuts *)
           var nsuc: integer;
           var ntry: Int64;
(* number of succesful and tried simulations *)
     var small, large: ResB;
(* indicate whether the minima and maxima, respectively,
   are smaller than the bounds set by 0 and m2 *)
     var s, s2, c: Res; var sp, sp2: real;
     var sl, s2l, cl: ResL; var ssl: ResL2
(* output: sums, sums of squares and of cross-products *)
 );

procedure ZOSim(g, mode: Byte;
(* First come input arguments : *)
(*         g = dimensions of the matrix *)
(*  mode   last digit is called basemode (1-7);
    mode = 2: square matrix with structurally zero diagonal (digraph);
           3: symmetric matrix with structurally zero diagonal (graph);
           5: as 2, with number of mutual dyads = symmetric (1,1) pairs
                    fixed at the value presmuts;
           6: as 2, restriction to connected digraphs;
           7: as 3, restriction to connected graphs;
           12,13,15,16,17: as 2,3,5,6,7 with rows and columns
              to be reordered according to decreasing values
              of |g - 1 - 2*rt[i]|.
           same + 100: Molloy-Reed algorithm (only if basemode 1,2,3)
    The advice is, always to use mode 12,13,15,16,17 and never 2,3,5,6,7. *)
           presmuts: integer;
(* see above *)
           var rt: Marginr; var ct: Marginc;
(* prescribed vectors of row sums and column sums *)
           var weit: TWeights;
(* weights for linear combinations of triad counts *)
           numl: Byte;
(* number of functions of triad counts evaluated *)
           var nrun: integer;
(* desired number of runs *)
(* Secondly output arguments : *)
           var errtext, warntext: shortstring;
(* if, after exiting from Simulate, errtext = '', then the procedure has
           proceeded correctly; if not, errtext indicates what went wrong;
           warntext can contain a warning *)
           var mutzero: integer;
(* origin for number of mutuals:
   distribution of first reported function in output Res arrays
   is distribution of mut minus mutzero;
   this transformation does not affect the prescribed value presmuts *)
           var meanp,varp: real; var mean,sd,cummean,cumsd: Res;
           var meanl,sdl: ResL; var covl: ResL2
(* means, standard errors, covariances *)
  );
(* Procedure ZOSim is a simplified version of procedure Simulate;
   the latter is contained in the implementation part of this unit. *)

implementation

uses SysUtils, ZOS_Form, FunctSim;

procedure unpackmode(var full,digr,graph,gnrl,
                         mutfix,rorder,connect,MolReed: boolean);
var ms: shortstring;
begin
  str(mode,ms);
   if (mode > 100) then begin
    MolReed := true;
    mode := mode-100;
   end else MolReed := false;
  if (mode < 1) then errtext := 'Mode '+ms+' less than 1; not implemented.';
  if ((mode in [6..11]) or (mode = 14) or (mode >= 16))
     then errtext := 'Mode '+ms+' not implemented.';
  full := (mode = 1);
  digr := mode in [2,5,6,12,15,16];
  graph := mode in [3,7,13,17];
  gnrl := (mode = 4);
  mutfix := ((mode = 5) or (mode = 15));
  rorder := (mode >= 12);
  connect := (mode in [6,7,16,17]);
  (* The following two checks are superfluous in the present implementation:
  if (Ord(full) + Ord(digr) + Ord(gnrl) + Ord (graph) < 1) then
    errtext := 'Number of requested methods is 0';
  if (Ord(full) + Ord(digr) + Ord(gnrl) + Ord (graph) > 1) then
    errtext := 'Number of requested methods is greater than 1'; *)
end; (* unpackmode *)

(* Some utilities : *)

function sumr(var a: Marginc; g: Byte): integer ;
var x: integer;
    k: Byte;
begin
 x := 0;
 for k:= 1 to g do x:= x + a[k];
 sumr := x;
end; (* sumr *)

function sumc(var a: Marginr; f: Byte): integer ;
var x: integer;
    k: Byte;
begin
 x := 0;
 for k:= 1 to f do x:= x + a[k];
 sumc := x;
end; (* sumc *)

function Root(x: Real): Real;
begin
  if x > 1e-36 then Root := sqrt(x) else Root := 0.0;
end; (* Root *)


function Connected(f,g: Byte; graph: boolean; mat: Adjac): Boolean;
(* Indicates whether mat is a connected (di)graph;
   if mat is not a graph, then it is symmetrized first. *)
(* should only be used for f=g *)
var i,j,k: byte; m,m2: Adjac; pow: Integer; conn: Boolean; ert: shortstring;
begin
  with m do begin
    AllocateMat(g,g,ert);
    Defequal(mat,ert);
    for i:=1 to g do Def(i,i,true);
    (* symmetrize : *)
    if (not graph) then for i:=1 to g do for j:=i-1 downto 1 do begin
      if (Arc(i,j) or Arc(j,i)) then begin
        Def(i,j,true); Def(j,i,true);
       end;
     end;
   end;
  with m2 do begin
    Allocatemat(g,g,ert);
    Defequal(m,ert);
   end;
  pow := 1;
  repeat
    with m do Defequal(m2,ert);
    pow := 2*pow;
    (* m2 := Boolean square of m : *)
    conn := true;
    for i:=g downto 1 do for j:=g downto 1 do begin
      k := 1;
      while ((not m2.Arc(i,j)) and (k <= g)) do begin
        if (m.ARc(i,k) and m.Arc(k,j)) then m2.Def(i,j,true);
        Inc(k);
       end; (* while *)
      conn := (conn and m2.Arc(i,j));
     end; (* for i and j *)
   until ((pow >= g) or conn);
  Connected := conn;
  m.Deallocatemat; m2.Deallocatemat;
end; (* Connected *)

(* Some informative non-essential procedures used in Simulate : *)

procedure Displays(nt1,nt2:integer);
begin
  ZOSimForm.showattempt(nt1);
  ZOSimForm.shownumber(nt2);
end; (* Displays *)

procedure HidePrelim;
begin
  ZOSimForm.PrelimLabel.visible := false;
end;

procedure Listmat(f,g: Byte; var mat: Adjac; nsuc: integer; np: real;
                  rff: rFuncvalue1; numf: Byte; digr: Boolean);
var i,j: Byte;
 d1,d2: Int64;
 frac: real;
begin
  if (nsuc mod 1000 = 0) then begin
    d1 := diskfree(0);
    d2 := disksize(0);
    frac := d1/d2;
    if (frac < 0.05) then begin
      writeln(outf,'You are writing all matrices to disk on the file ',pname,', ');
      writeln(outf,'but of the entire disk space only ',frac:4:2,' is free');
      writeln(outf,'after writing ',nsuc,' matrices.');
      writeln(outf,
      'In order to avoid filling up your entire disk space, writing file ',
      pname,' is terminated.');
      if (nsuc >= 10000) then writeln(outf,
        'You are advised to delete this file, it may be huge.');
      printall := false;
      printfileconn := false;
      closefile(pfile);
      exit;
     end;
   end; (* if *)
writeln(pfile);
writeln(pfile,nsuc,'.');
if digr then begin
  with mat do
  for i:=1 to g do begin
    for j:= 1 to i-1 do
    if Arc(i,j) then begin
      if Arc(j,i) then write(pfile,'M') else write(pfile,'1') end
     else write(pfile,'0');
    write(pfile,'X');
    for j:= i+1 to g do
    if Arc(i,j) then begin
      if Arc(j,i) then write(pfile,'M') else write(pfile,'1') end
     else write(pfile,'0');
    writeln(pfile);
   end; (* i *)
  end (* digr *) else begin
  with mat do
  for i:=1 to f do begin
    for j:= 1 to g do
      if Arc(i,j) then write(pfile,'1') else write(pfile,'0');
    writeln(pfile);
   end; (* i *)
  end; (* not digr *)
 for i:=1 to numf do begin
   if (i=1) then write(pfile,'@');
   write(pfile,' ',rff[i]:6:2);
   if ((i mod 12) = 0) then begin writeln(pfile); writeln(pfile,' '); end;
  end;
 write(pfile,np:15:8);
 writeln(pfile);
end; (* Listmat *)

(* now come the essential procedures for simulation : *)


procedure Initi(var ff: textfile; var errtext: shortstring;
     var numf, numl, numl2: Byte; var nsuc: integer;
     var ntry: Int64;
     var mutzer: integer;
     var small, large: ResB; var sl, s2l, cl: ResL; var ssl: ResL2;
     var s, s2, c: Res; var sp, sp2: real);
(* Initialisation for the bookkeeping procedures *)
var i: Byte; ii: integer;
 srr: shortstring;
begin
  randomize3(srr);
  writeln(ff,srr);
  halted := false; haltit := false;
  errtext := '';
  nsuc := 0; ntry := 0;
  mutzer := 0;
  for i:=1 to numfmax do begin
    small[i] := false;
    large[i] := false;
    for ii:= m1 to m2 do
      begin
        s[i,ii]:= 0.0;
        s2[i,ii] := 0.0;
        c[i,ii] := 0.0;
      end; (* ii *)
    end; (* i *)
  for i:=1 to numl2max do begin
    sl[i] := 0.0;
    s2l[i] := 0.0;
    cl[i] := 0.0;
    end; (* i *)
  for i:=1 to numlmax do
    for ii :=1 to numlmax do ssl[i,ii] := 0.0;
  sp:= 0.0;
  sp2 := 0.0;
end; (* initi *)

procedure Calcresult(nrun: integer; numf, numl, numl2: Byte;
(* input arrays and reals : *)
        var sp,sp2: real; var s,s2,c: Res;
        var sl,s2l,cl: ResL; var ssl: ResL2;
(* output arrays and reals : *)
        var meanp,varp: real; var mean,sd,cummean,cumsd: Res;
        var meanl,sdl: ResL; var covl: ResL2);
(* This procedure must follow Simulate, when the desired number of runs
   has been executed. *)
var ii: Integer;
    h,hh: Byte;
    covi, cums1, cums2, cumc: Real;
begin
  if (nrun <= 2) then exit;
(* Correspondence with formulae in Snijders, Psychometrika 1991, are given *)
  meanp := sp/nrun;            (* (1/n) Sum(r(X_i)) *)
  varp := (sp2/nrun - sqr(meanp))/(nrun-1);  (* (sigma^2_r) / n *)
  for h:=1 to numf do begin
    cums1 := 0.0;
    cums2 := 0.0;
    cumc  := 0.0;
    for ii:= m1 to m2 do
      begin
(* for phi = indicator function of Func[h] = ii : *)
        mean[h,ii] := s[h,ii]/nrun;          (* (1/n) Sum(phi(X_i)r(X_i)) *)
        covi  := (c[h,ii]/nrun  - meanp*mean[h,ii])/(nrun-1);
                            (* rho * sigma^2_r * sigma^2_phi,r / n *)
        if (ii >= 0) then begin
          cums1    := cums1 + mean[h,ii];
          cums2    := cums2 + s2[h,ii]/nrun;
          cumsd[h,ii] := (cums2 - sqr(cums1))/(nrun-1);
          cumc     := cumc + covi;
          cummean[h,ii] := cums1/meanp;
          cumsd[h,ii]   :=
 root(cumsd[h,ii] - 2*cumc*cummean[h,ii] + varp*sqr(cummean[h,ii]))/meanp;
         end; (* if ii *)
        sd[h,ii] := (s2[h,ii]/nrun - sqr(mean[h,ii]))/(nrun-1);
                                          (* sigma^2_phi,r / n *)
        mean[h,ii] := mean[h,ii]/meanp;   (* mu-hat *)
        sd[h,ii] :=
   root(sd[h,ii] - 2*covi*mean[h,ii] + varp*sqr(mean[h,ii]))/meanp ;
      end; (* ii *)
    mean[h,-2] := mean[h,-2] - sqr(mean[h,-1]);
    if (mean[h,-2] < 0.0) then mean[h,-2] := 0.0;
   end; (* h *)
  for h:=1 to numl2 do begin
(* for phi = fv2[h] : *)
    meanl[h] := sl[h]/nrun;          (* (1/n) Sum(phi(X_i)r(X_i)) *)
    covi  := (cl[h]/nrun  - meanp*meanl[h])/(nrun-1);
                        (* rho * sigma^2_r * sigma^2_phi,r / n *)
    sdl[h] := (s2l[h]/nrun - sqr(meanl[h]))/(nrun-1);
                                      (* sigma^2_phi,r / n *)
    meanl[h] := meanl[h]/meanp;   (* mu-hat *)
    sdl[h] :=
      root(sdl[h] - 2*covi*meanl[h] + varp*sqr(meanl[h]))/meanp ;
   end; (* h *)
  for h:=1 to numl do
    for hh:=1 to numl do
      covl[h,hh] := ssl[h,hh]/(nrun*meanp) - meanl[h]*meanl[hh];
end;      (* Calcresult *)

procedure Simulate(first: Boolean;
(* indicates whether this is the first of a simulation job,
   or continues a simulation job that has already started ; *)
           f,g: Byte;
(* dimensions of the matrix *)
           mode: Byte;
(* mode of matrix:
    mode = 1: matrix without structural zeroes;
           2: square matrix with structurally zero diagonal (digraph);
           3: symmetric matrix with structurally zero diagonal (graph);
           4: arbitrary pattern, structural zeroes in matrix strzer;
           5: as 2, with number of mutual dyads = symmetric (1,1) pairs
                    fixed at the value presmuts;
           6: as 2, restriction to connected digraphs;
           7: as 3, restriction to connected graphs;
           12,13,15,16,17: as 2,3,5,6,7 with rows and columns
              to be reordered according to decreasing values
              of |g - 1 - 2*rt[i]|
           same + 100: Molloy-Reed algorithm (only if basemode 1,2,3)   *)
           var strzer: Adjac; presmuts: Integer;
           var rt: Marginr; var ct: Marginc;
(* see above; for mode = 3,7,13 or 17,
   the value of the array ct has no input meaning
   and a graph with row and column sums as in rt will be simulated. *)
(* prescribed vectors of row sums and column sums *)
           var weit: TWeights;
(* weights for linear combinations of triad counts *)
           numl: Byte;
(* number of functions of triad counts evaluated *)
           var nrun: integer;
(* desired number of runs *)
           var errtext, warntext: shortstring;
(* if, after exiting from Simulate, errtext = '', then the procedure has
           proceeded correctly;if not, errtext indicates what went wrong;
           warntext can contain a warning *)
           var mutzero: integer;
(* origin for number of mutuals:
   distribution of first reported function in output Res arrays
   is distribution of mut minus mutzero;
   this transformation does not affect the prescribed value presmuts *)
           var nsuc: integer;
           var ntry: Int64;
(* number of succesful and tried simulations *)
     var small, large: ResB;
(* indicate whether the minima and maxima, respectively,
   are smaller than the bounds set by 0 and m2 *)
     var s, s2, c: Res; var sp, sp2: real;
     var sl, s2l, cl: ResL; var ssl: ResL2
(* output: sums, sums of squares and of cross-products *)
 );

var
  full, digr, gnrl, graph, mutfix, rorder, connect, MolReed: Boolean;
(* mode of matrix; of full, digr, gnrl, graph, exactly one is true;
           mutfix implies digr;
           full = matrix without structural zeroes;
           digr = square matrix with structurally zero diagonal (digraph);
           gnrl = arbitrary pattern, structural zeroes in matrix strzer;
           graph = symmetric matrix with structurally zero diagonal;
           mutfix = digr with number of mutual dyads = symmetric (1,1) pairs
                    fixed at the value presmuts;
           rorder = rows and columns are to be reordered;
                    only for graph or digr
           connect = restriction to connected graphs; only for graph       *)
  rav: Marginr;
  cav: Marginc;
    (* r refers to row, c to column;
       av refers to total number of available places, i.e.,
          entries which can be either "true" or "false"; *)
  r1,r0: Marginr;
  c1,c0: Marginc;
    (* number of available ones and zeroes; *)
  npra: Real;                 (* normalised inverse probability *)
  i: Integer;
  nul, asy, dyad: Integer;
                 (* constucted numbers of specific kinds of dyads *)
  n1x, nxx: Integer; (*  constructed 1x and xx dyads ;
                      x means that the entry is not yet determined *)
  mut: Integer;  (* constructed number of mutuals *)
  mat: Adjac;    (* constructed adjacency matrix *)
  z: Adjac;
   (* z[i,j] indicates that entry mat[i,j] is defined by the conditions;
      this matrix is called D in Snijders (Psychometrika, 1991) *)
  wrong: Boolean;
  impl00,impl10,impl11,impl01: Boolean;
      (* indicate for newly formed dyads that, for 00, a 0x dyad must
         become a 00 dyad, etc. ;
      the last 1 refers to the situation before the start of simulations *)

(* The following variables are defined once in procedure Initialise *)
  rav1: Marginr;
  cav1: Marginc;
    (* r refers to row, c to column;
       av refers to total number of available places, i.e.,
          entries which can be either "true" or "false";
       t  refers to the total number of entries "true" (marginal total);
       1 refers to the situation that holds for and before all simulations *)
  r11,r01: Marginr;
  c11,c01: Marginc;
    (* number of available ones and zeroes; the second 1 refers to the
       situation before all simulations *)
  rzer: Marginr; (* number of prescribed structural zeros in rows *)
  czer: Marginc;  (* the same in columns *)
  mut1, nul1, asy1, dyad1, n1x1, nxx1: Integer;
         (*  constructed 1x and xx dyads ;
                      x means that the entry is not yet determined *)
         (* the last 1 refers to the situation
            before the start of simulations *)
  z1: Adjac;
   (* z[i,j] indicates that entry mat[i,j] is defined by the conditions;
      this matrix is called D in Snijders (Psychometrika, 1991)
      the 1 refers to the situation before the start of the simulations *)
  impl001,impl101,impl111,impl011: Boolean;
      (* indicate for newly formed dyads that, for 00, a 0x dyad must
         become a 00 dyad, etc. ;
      the last 1 refers to the situation before the start of simulations *)
  presnuls, presasys, presdyads, mod1, mod2: Integer;
                 (* prescribed numbers of specific kinds of dyads *)
  ms: shortstring;

procedure Update(disf: Boolean; var mat: Adjac; npra: real;
                 mut: integer; numf, numl, numl2: Byte);
(* Called in Simulate after each succesful matrix construction *)
var ff: Integer; ff1: real;
    h,hh: Byte;
    x, xx, y, yy: real;
    fv1: Funcvalue1;
    fv2: Funcvalue2;
    rf1: RFuncvalue1;
begin
  x := npra;
  xx := sqr(x);
  sp  := sp + x;
  sp2 := sp2 + xx;
  Func(mat,f,g,numl,fversion,graph,digr,connect,mutzero,rt,weit,fv1,fv2,rf1);
  if printall then listmat(f,g,mat,nsuc,npra,rf1,numf,digr);
  if useresfile then writeln(resfile,x:12:6,' ');
  for h := 1 to numf do begin
    ff := fv1[h];
    if useresfile then begin
      write(resfile,ff:5,' ');
      if ((h mod 10) = 0) then writeln(resfile);
     end else (* not useresfile *) begin
      if (ff < 0) then begin
        ff := 0;
        small[h] := true;
       end; (* if *)
      if (ff <= m2) then begin
        s[h,ff]  := s[h,ff] + x;
        s2[h,ff] := s2[h,ff] + xx;
        c[h,ff]  := c[h,ff] + xx;
       end else large[h] := true;
      y  := ff*x;
      yy := ff*y;
      s[h,-1]  := s[h,-1] + y;
      s[h,-2]  := s[h,-2] + yy;
      s2[h,-1] := s2[h,-1] + sqr(y);
      s2[h,-2] := s2[h,-2] + sqr(yy);
      c[h,-1]  := c[h,-1] + y*x;
      c[h,-2]  := c[h,-2] + yy*x;
     end; (* not useresfile *)
   end; (* h *)
  if (useresfile and ((numf mod 10) > 0)) then writeln(resfile);
  for h:= numl2 downto 1 do begin
    ff1    := fv2[h];
    y     := ff1*x;
    sl[h] := sl[h] + y;
    s2l[h]:= s2l[h] + sqr(y);
    cl[h] := cl[h] + y*x;
   end; (* h *)
  for h:=numl downto 1 do
    for hh:= numl downto 1 do ssl[h,hh] := ssl[h,hh] + x*fv2[h]*fv2[hh];
end;     (* Update *)

function WrongMut: Boolean;
(* indicates that the number of mutuals cannot be equal to presmuts *)
begin
  WrongMut := ((mut > presmuts) or (nul > presnuls) or (asy > presasys)
              or (presmuts > mut + nxx + n1x) or
                 (presnuls > nul + presdyads - n1x - dyad));
end; (* WrongMut *)

procedure St0(i,j: Byte; var wrong: Boolean);
begin
  if (r0[i] > 0) then r0[i] := r0[i]-1 else
    begin wrong := true; Exit; end;
  if (c0[j] > 0) then c0[j] := c0[j]-1 else
    begin wrong := true; Exit; end;
  if digr then with z do if (Arc(j,i) and (i <> j)) then begin
     Inc(dyad);
    with mat do if Arc(j,i)
            then Inc(asy) else Inc(nul); end;
  if (mutfix and (i<>j)) then begin
    with z do if Arc(j,i) then begin
      with mat do if Arc(j,i) then n1x := n1x - 1; end
     else nxx := nxx - 1;
    if WrongMut then
      begin wrong := true; Exit; end;
   end; (* mutfix and i<>j *)
  with z do Def(i,j,true);
  with mat do Def(i,j,false);
  rav[i] := rav[i]-1;
  cav[j] := cav[j]-1;
end; (* St0 *)

procedure St1(i,j: Byte; var wrong: Boolean);
begin
  if (r1[i] > 0) then r1[i] := r1[i]-1 else
    begin wrong := true; Exit; end;
  if (c1[j] > 0) then c1[j] := c1[j]-1 else
    begin wrong := true; Exit; end;
  if digr then with z do if (Arc(j,i) and (i <> j)) then begin
    Inc(dyad);
    with mat do if Arc(j,i)
           then Inc(mut) else Inc(asy); end;
  if (mutfix and (i<>j)) then begin
    with z do if Arc(j,i) then begin
      with mat do if Arc(j,i) then n1x := n1x-1; end
     else begin
       nxx := nxx-1;
       Inc(n1x); end;
    if WrongMut then
      begin wrong := true; Exit; end;
   end; (* mutfix and i<>j *)
  with z do Def(i,j,true);
  with mat do Def(i,j,true);
  rav[i] := rav[i]-1;
  cav[j] := cav[j]-1;
end; (* St1 *)

procedure Set0(i,j: Byte; var wrong: Boolean);
begin
  St0(i,j,wrong);
  if wrong then Exit;
  if graph then begin
    with z do if (not Arc(j,i)) then St0(j,i,wrong);
    if wrong then Exit; end else
   if mutfix then begin
    with z do if (not Arc(j,i)) then
       begin if impl00 then St0(j,i,wrong)
          else if impl01 then St1(j,i,wrong); end;
    if wrong then Exit;
    end;
end; (* Set0 *)

procedure Set1(i,j: Byte; var wrong: Boolean);
begin
  St1(i,j,wrong);
  if wrong then Exit;
  if graph then begin
    with z do if (not Arc(j,i)) then St1(j,i,wrong);
    if wrong then Exit; end else
   if mutfix then begin
     with z do if (not Arc(j,i)) then
       begin if impl10 then St0(j,i,wrong)
         else if impl11 then St1(j,i,wrong); end;
    if wrong then Exit; end;
end; (* Set1 *)

procedure S0(i,j: Byte; var c: Boolean; var wrong: Boolean);
begin Set0(i,j,wrong); c := true; end;

procedure S1(i,j: Byte; var c: Boolean; var wrong: Boolean);
begin Set1(i,j,wrong); c := true; end;

procedure CheckRem(i1,i2,j1,j2: Byte; var wrong, change: Boolean);
(* This procedure checks if the remainder of rows i1 to i2
   or columns j1 to j2 are defined
   and, if so, carries out the implied assignments *)
var rr1,rc1,rr0,rc0: Integer;
    jj,i,j: Byte;

begin
  wrong := false; change := false;
  for i := i1 to i2 do begin
    rr1 := r1[i];        (* entries 1 for remainder row i *)
    rr0 := r0[i];        (* entries 0 for remainder row i *)
        if (rr1 <= 0) then begin
      for jj:= g downto 1 do
        begin with z do if (not Arc(i,jj)) then S0(i,jj,change,wrong);
          if wrong then Exit; end;
       end else if (rr0 <= 0) then begin
      for jj:= g downto 1 do
        begin with z do if (not Arc(i,jj)) then S1(i,jj,change,wrong);
          if wrong then Exit; end;
       end;
  end; (* i *)
  for j := j1 to j2 do begin
    rc1 := c1[j];        (* entries 1 for remainder column j *)
    rc0 := c0[j];        (* entries 0 for remainder column j *)
    if (rc1 <= 0) then begin
      for jj:= f downto 1 do
        begin with z do if (not Arc(jj,j)) then S0(jj,j,change,wrong);
          if wrong then Exit; end;
       end else if (rc0 <= 0) then begin
      for jj:= f downto 1 do
        begin with z do if (not Arc(jj,j)) then S1(jj,j,change,wrong);
          if wrong then Exit; end;
       end;
  end; (* j *)
end;  (* CheckRem *)

procedure Checkmut(istar: Byte; var wrong, change: Boolean);
(* This procedure is called only in the MutFix mode,
   and carries out assignments that follow from the conditions
   where the further created number of mutuals is minimal
   or maximal, respectively *)
var i,j: Byte;

begin
  change := false;
  wrong := WrongMut;
  if wrong then Exit;
  if ((not impl10) and (mut >= presmuts)) then begin
  (* no more mutuals may be formed *)
     for i:= istar to g do
       for j :=g downto 1 do
         with z do if ((not Arc(i,j)) and Arc(j,i)) then
           begin with mat do if Arc(j,i) then S0(i,j,change,wrong);
             if Wrong then Exit; end;
     impl10 := true;
    end;
  if ((not impl01) and (nul >= presnuls)) then begin
  (* no more nulls may be formed *)
     for i:= istar to g do
       for j :=g downto 1 do
         with z do if ((not Arc(i,j)) and Arc(j,i)) then
       begin with mat do if (not Arc(j,i)) then S1(i,j,change,wrong);
        if Wrong then Exit; end;
     impl01 := true;
    end;
  if ((not impl00) and (presnuls >= nul + presdyads - dyad - n1x)) then begin
   (* all new entries 1 must be put in 1x dyads,
        and all 0x and xx dyads must become 00 dyads  *)
     for i:= istar to g do
       for j :=g downto 1 do
         with z do if (not Arc(i,j)) then begin
          if Arc(j,i) then
   begin with mat do if (not Arc(j,i)) then S0(i,j,change,wrong);
         if Wrong then Exit; end
           else begin S0(i,j,change,wrong);
                if (not wrong) then S0(j,i,change,wrong);
                if Wrong then Exit; end;
          end;
     impl00 := true;
    end;
  if ((not (impl00 and impl11)) and (asy >= presasys)) then begin
(* all 1x dyads must become mutuals and all 0x dyads must become 00 dyads *)
    for i:= istar to g do
      for j :=g downto 1 do
        with z do if ((not Arc(i,j)) and Arc(j,i)) then
          begin with mat do if Arc(j,i) then S1(i,j,change,wrong)
                 else S0(i,j,change,wrong);
            if Wrong then Exit; end;
    impl00 := true; impl11 := true;
   end;
  if ((not impl11) and (presmuts >= (mut + n1x + nxx))) then
    begin
        (* all 1x and xx dyads must become mutuals *)
    for i:= istar to g do
      for j :=g downto 1 do
        with z do if (not Arc(i,j)) then begin
          if Arc(j,i) then
            begin with mat do if Arc(j,i) then S1(i,j,change,wrong);
                  if Wrong then Exit; end
                else begin S1(i,j,change,wrong);
                 if (not wrong) then S1(j,i,change,wrong);
                 if Wrong then Exit; end;
         end;
    impl11 := true;
   end;
end; (* CheckMut *)

procedure Order(var r: Marginr);
(* gives as output in r the increasing rank orders of fff(i) *)
function fff(i: Byte): Integer;
begin
  if ((i >= 1) and (i <= g)) then
   fff := -abs((g-1)-2*rt[i])
  else fff := 0;
end; (* fff *)
var i,j: Byte;
begin
  for i:=1 to g do r[i] := 1;
  for i:= 2 to g do
    for j:=1 to i-1 do if (fff(j) > fff(i)) then Inc(r[j]) else Inc(r[i]);
end; (* Order *)

procedure Reorder(var rt: Marginr; var ct: Marginc);
(* reorder rows and columns; only for f=g! *)
var i: Byte;
    r,r0: Marginr; c0: Marginc;
begin
  r := rt;
  Order(r);
  ChangedOrder := false;
  for i:=1 to g do begin
    if (r[i] <> i) then ChangedOrder := true;
    r0[r[i]] := rt[i];
    c0[r[i]] := ct[i];
   end;
  rt := r0;
  ct := c0;
  newone := r[1];
  newtwo := r[2];
  if ChangedOrder then begin
    writeln(outf,'For efficiency purposes, rows and columns are similarly reordered.');
    writeln(outf,'Correspondence between old and new vertex numbers :');
    write(outf,'old  : ');
    if (f < 100) then begin
      for i:=1 to f do write(outf,i:3); writeln(outf);
     end else begin
      for i:=1 to f do write(outf,i:4); writeln(outf);
     end;
    write(outf,'new  : ');
    if (f < 100) then begin
      for i:=1 to f do write(outf,r[i]:3); writeln(outf);
     end else begin
      for i:=1 to f do write(outf,r[i]:4); writeln(outf);
     end;
    writeln(outf,'Look below at the row and column sums to check the new order.');
   end;
end; (* Reorder *)

procedure Initialise(var Errtext: shortstring; var resf: Text);
(* initialisation for simulations *)
var i,j: Byte;
    gint: Integer;
    change, ch: Boolean;
    fna: shortstring;
begin
  Errtext := '';
  numf := NumFunc(fversion,f,g);
  if (numf > numfmax) then begin
    if (numf > numfplus) then begin
      Errtext := 'numf = '+inttostr(numf)+' is larger than numfplus = '+inttostr(numfplus);
      haltit := true;
      exit;
     end; (* if *)
    fna := proname+'.pqr';
    writeln(outf,'(Number of statistics evaluated is larger than constant Numfmax.');
    writeln(outf,'Therefore, intermediate results will be written to file ',fna,').');
    assign(resf,fna);
    rewrite(resf);
    writeln(resf,numf);
    for i:= 1 to numf do writeln(resf,Tex(i,mutzero,fversion,graph,digr,connect,true));
    writeln(resf); writeln(resf,'@');
    useresfile := true;
   end (* numf > numfmax *) else useresfile := false;
  if (numl > numlmax) then errtext := 'Numl is larger than NumlMax.';
  numl2 := 2*numl;
  with mat do AllocateMat(f,g,errtext);
  if (errtext > '') then exit;
  with z do AllocateMat(f,g,errtext);
  if (errtext > '') then exit;
  with z1 do AllocateMat(f,g,errtext);
  if (errtext > '') then exit;
  if graph then
    for i:=1 to g do ct[i] := rt[i];
  for i:=1 to f do rzer[i] := 0;
  for j:=1 to g do czer[j] := 0;
  for i:= 1 to f do begin
    for j:= 1 to g do begin
      with mat do Def(i,j,false);
      with z1 do Def(i,j,false);
      with strzer do if Arc(i,j) then begin Inc(rzer[i]); Inc(czer[j]); end;
     end; (* j *)
    with z do Defequal(strzer,errtext);
   end; (* i *)
  for i:=1 to f do if ((rt[i] < 0) or (rt[i] + rzer[i] > g)) then begin
    Errtext := 'A row total is impossible.'; Exit; end;
  for j:=1 to g do if ((ct[j] < 0) or (ct[j] + czer[j] > f)) then begin
    Errtext := 'A column total is impossible.'; Exit; end;
  for i:= 1 to f do begin
    r0[i] := g-rzer[i]-rt[i];
    r1[i]   := rt[i];
    rav[i] := g - rzer[i];
   end; (* i *)
  for j:= 1 to g do begin
    c0[j] := f-czer[j]-ct[j];
    c1[j]   := ct[j];
    cav[j] := f - czer[j];
   end; (* j *)
  mut := 0; nul := 0; asy := 0; dyad := 0;
  if (Graph or MutFix) then begin
    n1x := 0;  gint := g; nxx := (gint*(gint-1)) div 2;
    presnuls := nxx + presmuts - sumc(rt,f);
    presasys := nxx - presmuts - presnuls;
    presdyads := nxx;
   end;
  impl00 := false; impl10 := false; impl11 := false; impl01 := false;
  if graph then begin
    impl00 := true; impl11 := true; end;
 (* The following is to see if the row and column totals already determine
    part of mat. *)
   repeat
     CheckRem(1,f,1,g,wrong,change);
     if (mutfix and (not wrong)) then Checkmut(1,wrong,ch) else ch := false;
     if wrong then begin
 Errtext := 'There is no adjacency matrix subject to these restrictions.';
       exit;
       end;
    until (not (change or ch));
  with z1 do Defequal(z,errtext);
  r01 := r0;
  r11 := r1;
  c01 := c0;
  c11 := c1;
  rav1 := rav;
  cav1 := cav;
  mut1 := mut;
  nul1 := nul;
  asy1 := asy;
  dyad1 := dyad;
  if mutfix then begin n1x1 := n1x; nxx1 := nxx; end;
  impl001 := impl00;
  impl101 := impl10;
  impl111 := impl11;
  impl011 := impl01;
end; (* Initialise *)

function pprr(x1, n1, x2, n2, x3, n3: Integer; var wrong: Boolean):Real;
(* Conditional probability of a 1 in the currently determined cell *)
(* no check is made for 0 <= xi <= ni etc. *)
var p1,p2,p1p2,p3,nn: Real;
begin
  p1 := x1/n1;
  p2 := x2/n2;
  p3 := (x3-x1+1)/(n3-n1+2);
  p1p2 := p1*p2;
  nn := p1p2 + p3*(1-p1-p2);
  if (nn > 1e-12) then
    begin pprr := p1p2*(1-p3) / nn; wrong := false; end else
    begin pprr := 0.0; wrong := true; end;
end; (* pprr *)


procedure Generate(starter: Boolean);
(* starter indicates that the generation is for normalization purposes
   at the start of the simulations *)
(* Correspondence with formulas in Snijders (Psychometrika 1991):
   r1[i] = r[i] - X[i,+](D);  rav[i] = g - D[i,+];
   c1[j]  = c[j] - X[+,j](D);  cav[j] = f - D[+,j];
   sijd = S[i,j] + r[i] - X[i,+](D);
   nijd = N[i,j] + g - D[i,+] *)
label Newstart;
var  i, ii, j: Byte; sijd, nijd: Integer;
     pra, iprob, ipr: real;
     change, ch: Boolean;
     erotext: shortstring;
     a: Marginr;
     cumb: array[1..gmax] of integer;
begin
Newstart:
  ZOSimForm.ProcessThem;
  if (((nsuc < 3) and (ntry mod 10 = 0)) or (ntry mod 100 = 0))
    then Displays(ntry,nsuc);
  if halted then exit;
  if starter then iprob := 0.0 else iprob := 1.0;
  Inc(ntry);
  with z do DefEqual(z1,erotext);
  r0 := r01;
  r1 := r11;
  c0 := c01;
  c1 := c11;
  rav := rav1;
  cav := cav1;
  mut := mut1;
  nul := nul1;
  asy := asy1;
  dyad := dyad1;
  if mutfix then begin n1x := n1x1; nxx := nxx1; end;
  if (graph or mutfix) then begin
    impl00 := impl001;
    impl10 := impl101;
    impl11 := impl111;
    impl01 := impl011;
   end;
  if MolReed then begin
(* generate a random contingency table,
   for "graph" a symmetric contingency table,
   with the given row and column sums;
   repeat this until the contingency table is a permitted adjacency matrix. *)
    a := rt;
    cumb[1] := ct[1];
    for j:=2 to g do cumb[j] := cumb[j-1]+ct[j];
    for i:=1 to f do begin
      while a[i] >0 do begin
        a[i] := a[i]-1;
        if graph then begin
          for j:=i to g do cumb[j] := cumb[j]-1;
         end;
        pra := random3*cumb[g];
        j := 1;
        while ((cumb[j] <= pra) and (j < g)) do inc(j);
        if (z.Arc(i,j) or strzer.Arc(i,j)) then goto NewStart;
        Set1(i,j,wrong);
        if wrong then goto NewStart;
        for ii:=j to g do cumb[ii] := cumb[ii]-1;
        if graph then a[j] := a[j]-1;
       end; (* while a[i] > 0 *)
     end; (* for i *)
    for i:=1 to f do for j:=1 to g do
      if not z.Arc(i,j) then Set0(i,j,wrong);
    if wrong then goto NewStart;
    ipr := 1.0;
   end else begin (* not MolReed *)
    for i:= 1 to f-1 do begin
      ipr := 1.0;
      for j:= g downto 2 do begin
        repeat
          CheckRem(i,f,1,g,wrong,change);
          if wrong then goto NewStart;
          if mutfix then begin
            CheckMut(i,wrong,ch); if wrong then goto NewStart; end
            else ch := false;
         until (not (change or ch));
        with z do if Arc(i,j) then
          (* mat[i,j] is already determined and nothing needs to be done *)
         else begin
          nijd := 0; sijd := 0;
          for ii := j-1 downto 1 do
            with z do if (not Arc(i,ii)) then begin
              Inc(nijd, cav[ii]);
              Inc(sijd, c1[ii]);
             end;
          pra := pprr(r1[i],rav[i],c1[j],cav[j],sijd,nijd,wrong);
          if wrong then goto NewStart;
          if random3 <= pra then
            begin
              Set1(i,j,wrong);
              if wrong then goto Newstart;
              ipr := ipr / pra;
             end else
            begin
              Set0(i,j,wrong);
              if wrong then goto NewStart;
              ipr := ipr / (1.0 - pra);
             end; (* if random3 *)
           end; (* not z[i,j] *)
         end; (* j *)
        with z do if (not Arc(i,1)) then begin
          CheckRem(i,i,1,1,wrong,change);
          if wrong then goto NewStart;
         end;
        repeat
          CheckRem(i+1,f,1,g,wrong,change);
          if wrong then goto NewStart;
          if mutfix then begin
            CheckMut(i+1,wrong,ch); if wrong then goto NewStart; end
            else ch := false;
         until (not (change or ch));
      if starter then
        iprob := iprob + ln(ipr)
         else iprob := ipr*iprob/normfac;
     end; (* i *)
   end; (* if not MolReed *)
  if connect then if (not(Connected(f,g,graph,mat))) then goto NewStart;
  if starter then begin
    normfac := normfac + iprob;
    Displays(ntry,nsuc);
   end;
  npra := iprob;
  Inc(nsuc);
  if not starter then Update(true,mat,npra,mut,numf,numl,numl2);
end;    (* procedure Generate *)

procedure Findnorm(var normfac: real; var mutzero: integer;
                   var nrun, nsuc: integer;
                   var ntry: Int64; var errtext: shortstring);
(* determines a normalizing factor for the probabilities *)
var ii: Byte; mutmin, mutmax: integer;
begin
  ZOSimForm.ProcessThem;
  if halted then exit;
  errtext := '';
  normfac := 0.0;
  mutmin := sumc(rt,f) div 2;
  mutmax := 0;
  for ii := 1 to 10 do
    begin Generate(true);
          ZOSimForm.ProcessThem;
          if halted then errtext := 'Program interrupted by user.';
          if (errtext > '') then exit;
          if (mut > mutmax) then mutmax := mut;
          if (mut < mutmin) then mutmin := mut;
         end;
  HidePrelim;
  normfac := exp(normfac/(10*(f-1)));
(* The following calculation of a lower bound to the number of mutuals
   is usually satisfactory, but it has a random nature.
   Incidentally it may happen that the number of mutuals realized in the
   simulation runs is not contained between mutzero and mutzero + m2.
   This will then appear in values TRUE for large[1] and/or small[1]. *)
  mutzero := 2*mutmin - mutmax - 10;
  if (mutzero < 0) then mutzero := 0;
  mutzero := 10*(mutzero div 10);
  nsuc := 0;
  ntry := 0;
end; (* Findnorm *)

begin (* procedure Simulate *)
errtext := ''; warntext := '';
if (f + g < 20) then begin
  mod1 := 100; mod2 := 1000;
 end else begin
  mod1 := 100; mod2 := 100;
 end;
if (f + g > 50) then begin
  mod1 := 10; mod2 := 10;
 end;
if (f + g > 100) then begin
  mod1 := 1; mod2 := 1;
 end;
if (f > fmax) then
  errtext := 'Number of rows f greater than maximum fmax.';
if (g > gmax) then
  errtext := 'Number of columns g greater than maximum gmax.';
unpackmode(full,digr,graph,gnrl,mutfix,rorder,connect,MolReed);
if (errtext > '') then begin
  haltit := true;
  exit;
 end;
if (sumr(ct,g) <> sumc(rt,f)) then begin
  errtext := 'Rows and columns do not have the same sum.';
 end;
mut := sumr(ct,g);
if (mut <= 0) then
  errtext := 'All column sums are zero. trivial result!';
if (graph and (2*(mut div 2) < mut)) then
  errtext := 'Degree sum of a graph should be even!';
if mutfix then begin
  if (not digr) then
    errtext := 'Number of mutuals is fixed without request for a digraph.';
  if (presmuts < 0) then
    errtext := 'Prescribed number of mutuals is negative.';
  if (2*presmuts > mut) then
    errtext := 'Twice the prescribed number of mutuals is larger than number of ones.';
 end;
if (numl > numlmax) then errtext :=
  'Number of linear combinations of triad counts greater than constant numlmax.';
if (errtext > '') then begin
  haltit := true;
  exit;
 end;

if first then begin
  if rorder then Reorder(rt,ct);
  Initi(outf,errtext, numf, numl, numl2, nsuc, ntry, mutzero,
        small, large, sl, s2l, cl, ssl, s, s2, c, sp, sp2);
  if (errtext > '') then begin
    haltit := true;
    exit;
   end;
 end; (* if first *)

Initialise(errtext,resfile);
if (errtext > '') then begin
  haltit := true;
  exit;
 end;

if first then begin
  Findnorm(Normfac, mutzero, nrun, nsuc, ntry, errtext);
  if (errtext > '') then begin
    haltit := true;
    exit;
   end;
 end (* if first *) else
   if useresfile then append(resfile);

if printall then begin
(* only now, mutzero has been well defined *)
  if ChangedOrder then begin
    writeln(pfile,'The vertex numbers for the statistics refer to the numbers in the');
    writeln(pfile,'original order; consult the main output file for the reordering.');
    writeln(pfile,'The original vertex 1 now is ',newone,
                  '; the original vertex 2 now is ',newtwo,'.');
   end;
  writeln(pfile,'Following each matrix are the following function values:');
  for i:=1 to numf do writeln(pfile,i:2,'. ',
                Tex(i,mutzero,fversion,graph,digr,connect,false));
  writeln(pfile);
  writeln(pfile,'After this follows a constant divided by the probability');
  writeln(pfile,'for obtaining this matrix.');
  writeln(pfile,'For unbiased estimation, weights must be used');
  writeln(pfile,'that are proportional to these numbers.');
 end;

while (nsuc < nrun) do begin
   Generate(false);
   if ((nsuc < 100) or ((nsuc < mod2) and (nsuc mod mod1 = 0))
         or (nsuc mod mod2 = 0) or (ntry mod mod2 = 0)) then
               displays(ntry,nsuc);
    ZOSimForm.ProcessThem;
    if halted then nrun := nsuc;
 end;
displays(ntry,nsuc);
ZOSimForm.ProcessThem;
if small[1] then begin
  str(mutzero,warntext);
  warntext :=
  'Distribution of number of mutuals truncated at the left at the value '
                                                + warntext + ' .';
 end;
if large[1] then begin
  str(mutzero+100,warntext);
  warntext :=
  'Distribution of number of mutuals truncated at the right at the value '
                                                + warntext + ' .';
 end;
if (large[1] and small[1]) then
  warntext :=
  'Distribution of number of mutuals truncated at left and right. ';
with mat do DeallocateMat;
with z do DeallocateMat;
with z1 do DeallocateMat;
if useresfile then closefile(resfile);
end; (* Simulate *)

procedure ZOSim(g, mode: Byte;
(*         g = dimensions of the matrix *)
(*  mode = 2: square matrix with structurally zero diagonal (digraph);
           3: symmetric matrix with structurally zero diagonal (graph);
           5: as 2, with number of mutual dyads = symmetric (1,1) pairs
                    fixed at the value presmuts;
           6: as 2, restriction to connected digraphs;
           7: as 3, restriction to connected graphs;
           12,13,15,16,17: as 2,3,5,6,7 with rows and columns
              to be reordered according to decreasing values
              of |g - 1 - 2*rt[i]| *)
           presmuts: integer;
           var rt: Marginr; var ct: Marginc;
(* prescribed vectors of row sums and column sums *)
           var weit: TWeights;
(* weights for linear combinations of triad counts *)
           numl: Byte;
(* number of functions of triad counts evaluated *)
           var nrun: integer;
(* desired number of runs *)
           var errtext, warntext: shortstring;
(* if, after exiting from ZOSim, errtext = '', then the procedure has
           proceeded correctly; if not, errtext indicates what went wrong;
           warntext can contain a warning *)
           var mutzero: integer;
           var meanp,varp: real; var mean,sd,cummean,cumsd: Res;
           var meanl,sdl: ResL; var covl: ResL2
(* final results *)
  );

(* This procedure is a simplified version of procedure Simulate:
   it does only graphs and digraphs; structural zeros only on diagonal;
   no possibility to resume simulations after achieving nrun runs. *)
var zof: Byte;
    zostrzer: Adjac;
    zonsuc: integer;
    zontry: Int64;
    zosmall, zolarge: ResB;
    zos,zos2,zoc: Res;
    zosp,zosp2: real;
    zosl,zos2l,zocl: ResL;
    zossl: ResL2;
    i,j: Byte;
begin
  zof := g;
  with zostrzer do AllocateMat(zof,g,errtext);
  if (errtext > '') then begin
    haltit := true;
    exit;
   end;
  with zostrzer do begin
    for i:= 1 to zof do
      for j:=1 to g do if (i = j) then Def(i,i,true) else Def(i,j,false);
   end;
  Simulate(true,zof,g,mode,zostrzer,presmuts,rt,ct,weit,numl,nrun,
           errtext,warntext,mutzero,zonsuc,zontry,
           zosmall,zolarge,
           zos,zos2,zoc,zosp,zosp2,zosl,zos2l,zocl,zossl);
  if ((errtext > '') and (not halted)) then haltit := true;
  if (not haltit) then Calcresult(nrun, numf, numl, numl2,
        zosp,zosp2,zos,zos2,zoc,zosl,zos2l,zocl,zossl,
        meanp,varp, mean,sd,cummean,cumsd, meanl,sdl,covl);
  with zostrzer do DeallocateMat;
end; (* ZOSim *)

begin (* unit Simul *)
  newone := 1;
  newtwo := 2;
  ChangedOrder := false;
  OrderDependent := false;
end.