                           (* TRIADS.PAS *)


(* Collection of types and procedures for calculating the triad census
   of a directed graph.
   Programmed by Tom Snijders, University of Groningen *)

unit Triads;

interface

uses ZOMat;

type
  Three = 0..3;
  Matrix = array[1..30,1..30] of Byte;
  TriadVector = array[1..16] of Real;
  TriadType = array[1..3] of Three;
    (* Triads indicated in the Holland-Leinhardt notation by
       MAN or MANX, where M,A,N are numbers 0,1,2,3 and X is C,T,U, or D,
       are coded here by M,A,X since 2*M+A=3.
       Codes for X are: nothing = 0, C = 0, T = 1, U = 1, D = 2 *)
  TriadCountType = array[1..16] of integer;
  TriadArrayType = array[1..16] of TriadType;

function TriadName(var tr: TriadType): string;
function TriadNumber(var tr: TriadType): Byte;
procedure DetermineTriad(var m: Adjac; i,j,k: Byte; var t: TriadType);
procedure TriadCount(var m: Adjac; g: Byte; var c: triadcounttype);
procedure TriadProportion(var m: Adjac; g: Byte; var p: TriadVector);
var TriadKind: TriadArrayType;

implementation

procedure InitTriad(var Triad: TriadArrayType);
var t: TriadType;
begin
  t[1] := 0; t[2] := 0; t[3] := 0; Triad[1] := t;
  t[1] := 0; t[2] := 1; t[3] := 0; Triad[2] := t;
  t[1] := 1; t[2] := 0; t[3] := 0; Triad[3] := t;
  t[1] := 0; t[2] := 2; t[3] := 2; Triad[4] := t;
  t[1] := 0; t[2] := 2; t[3] := 1; Triad[5] := t;
  t[1] := 0; t[2] := 2; t[3] := 0; Triad[6] := t;
  t[1] := 1; t[2] := 1; t[3] := 2; Triad[7] := t;
  t[1] := 1; t[2] := 1; t[3] := 1; Triad[8] := t;
  t[1] := 0; t[2] := 3; t[3] := 1; Triad[9] := t;
  t[1] := 0; t[2] := 3; t[3] := 0; Triad[10] := t;
  t[1] := 2; t[2] := 0; t[3] := 0; Triad[11] := t;
  t[1] := 1; t[2] := 2; t[3] := 2; Triad[12] := t;
  t[1] := 1; t[2] := 2; t[3] := 1; Triad[13] := t;
  t[1] := 1; t[2] := 2; t[3] := 0; Triad[14] := t;
  t[1] := 2; t[2] := 1; t[3] := 0; Triad[15] := t;
  t[1] := 3; t[2] := 0; t[3] := 0; Triad[16] := t;
end;

function TriadName(var tr: TriadType): string;
begin
  case tr[1] of
    0:  case tr[2] of 0: TriadName := '003 ';
                      1: TriadName := '012 ';
                      2: case tr[3] of 0: TriadName := '021C';
                                       1: TriadName := '021U';
                                       2: TriadName := '021D';
                                       else TriadName := 'XXXX';
                         end;
                      3: case tr[3] of 0: TriadName := '030C';
                                       1: TriadName := '030T';
                                       else TriadName := 'XXXX';
                         end;
                      else TriadName := 'XXXX';
        end;
    1:  case tr[2] of 0: TriadName := '102 ';
                      1: case tr[3] of 1: TriadName := '111U';
                                       2: TriadName := '111D';
                                       else TriadName := 'XXXX';
                         end;
                      2: case tr[3] of 0: TriadName := '120C';
                                       1: TriadName := '120U';
                                       2: TriadName := '120D';
                                       else TriadName := 'XXXX';
                         end;
                      else TriadName := 'XXXX';
        end;
    2:  case tr[2] of 0: TriadName := '201 ';
                      1: TriadName := '210 ';
                      else TriadName := 'XXXX';
        end;
    3:  if ((tr[2] = 0) and (tr[3] = 0)) then TriadName := '300 '
                                         else TriadName := 'XXXX';
   end; (* case tr[1] *)
end; (* TriadName *)

function TriadNumber(var tr: TriadType): Byte;
begin
  case tr[1] of
    0:  case tr[2] of 0: TriadNumber := 1;
                      1: TriadNumber := 2;
                      2: case tr[3] of 0: TriadNumber := 6;
                                       1: TriadNumber := 5;
                                       2: TriadNumber := 4;
                                       else TriadNumber := 0;
                         end;
                      3: case tr[3] of 0: TriadNumber := 10;
                                       1: TriadNumber := 9;
                                       else TriadNumber := 0;
                         end;
                      else TriadNumber := 0;
        end;
    1:  case tr[2] of 0: TriadNumber := 3;
                      1: case tr[3] of 1: TriadNumber := 8;
                                       2: TriadNumber := 7;
                                       else TriadNumber := 0;
                         end;
                      2: case tr[3] of 0: TriadNumber := 14;
                                       1: TriadNumber := 13;
                                       2: TriadNumber := 12;
                                       else TriadNumber := 0;
                         end;
                      else TriadNumber := 0;
        end;
    2:  case tr[2] of 0: TriadNumber := 11;
                      1: TriadNumber := 15;
                      else TriadNumber := 0;
        end;
    3:  if ((tr[2] = 0) and (tr[3] = 0)) then TriadNumber := 16 else TriadNumber := 0;
   end; (* case tr[1] *)
end; (* TriadNumber *)

procedure DetermineTriad(var m: Adjac; i,j,k: Byte; var t: TriadType);
var mut,sum: Byte;
function Cycle: boolean;
begin
  with m do
  Cycle := (((Arc(i,j)) and (Arc(j,k)) and (Arc(k,i)))
            or
            ((Arc(j,i)) and (Arc(i,k)) and (Arc(k,j))));
  end; (* Cycle *)
function TwoIn: Boolean;
(* At least one in-degree 2. *)
  begin
  with m do
  TwoIn := (((Arc(i,j)) and (Arc(k,j))) or
            ((Arc(i,k)) and (Arc(j,k))) or
            ((Arc(j,i)) and (Arc(k,i))));
  end; (* TwoIn *)
function TwoOut: Boolean;
(* At least one out-degree 2. *)
  begin
  with m do
  TwoOut := (((Arc(j,i)) and (Arc(j,k))) or
             ((Arc(k,i)) and (Arc(k,j))) or
             ((Arc(i,j)) and (Arc(i,k))));
  end; (* TwoOut *)
function ZeroIn: Boolean;
(* At least one in-degree 0. *)
  begin
  with m do
  ZeroIn :=(((not Arc(i,j)) and (not Arc(k,j))) or
            ((not Arc(i,k)) and (not Arc(j,k))) or
            ((not Arc(j,i)) and (not Arc(k,i))));
  end; (* ZeroIn *)

begin (* DetermineTriad *)
mut := 0; sum := 0;
with m do begin
  if (Arc(i,j)) then begin
   Inc(sum);
   if (Arc(j,i)) then
     begin Inc(mut); Inc(sum); end;
  end else if (Arc(j,i)) then Inc(sum);
 if (Arc(i,k)) then begin
   Inc(sum);
   if (Arc(k,i)) then
     begin Inc(mut); Inc(sum); end;
  end else if (Arc(k,i)) then Inc(sum);
 if (Arc(j,k)) then begin
   Inc(sum);
   if (Arc(k,j)) then
     begin Inc(mut); Inc(sum); end;
  end else if (Arc(k,j)) then Inc(sum);
 end; (* with m *)
t[1] := mut;
t[2] := sum - 2*mut;
t[3] := 0;
(* Because of the last assignment, only the triads MANX with X = T, U, or D
   need to have a further t[3] assignment. *)
 if (mut <= 0) then begin (*A1*)
  if (sum >= 3) then
    (* triadtype MAN = 030 *)
     begin if not Cycle then t[3] := 1; end
    else begin (*B*)
     if (sum >= 2) then begin (*C*)
       (* triadtype MAN = 021 *)
        if TwoIn then t[3] := 1 else
          begin if TwoOut then t[3] := 2; end;
        end (*C*) ; end (*B*) ;
    end (*A1*) else begin (*A2*)
      if (mut <= 1) then begin (*B*)
        if (sum >= 4) then begin (*C1*)
          (* triadtype MAN = 120 *)
         if not Cycle then
          begin if ZeroIn then t[3] := 2 else t[3] := 1; end;
         end (*C1*) else begin (*C2*)
           if (sum >= 3 ) then
            (* triadtype MAN = 111 *)
            begin if TwoOut then t[3] := 1 else t[3] := 2; end;
           end (*C2*); end (*B*); end (*A2*);
end; (* DetermineTriad *)

procedure TriadCount(var m: Adjac; g: Byte; var c: triadcounttype);
var i,j,k: Byte;
    t: TriadType;
    mut,sum: Byte;
    cc: array[0..3,0..3,0..3] of Word;
    tf: textfile;
function Cycle: boolean;
  begin
  with m do
  Cycle := (((Arc(i,j)) and (Arc(j,k)) and (Arc(k,i)))
            or
            ((Arc(j,i)) and (Arc(i,k)) and (Arc(k,j))));
  end; (* Cycle *)
function TwoIn: Boolean;
(* At least one in-degree 2. *)
  begin
  with m do
  TwoIn := (((Arc(i,j)) and (Arc(k,j))) or
            ((Arc(i,k)) and (Arc(j,k))) or
            ((Arc(j,i)) and (Arc(k,i))));
  end; (* TwoIn *)
function TwoOut: Boolean;
(* At least one out-degree 2. *)
  begin
  with m do
  TwoOut := (((Arc(j,i)) and (Arc(j,k))) or
             ((Arc(k,i)) and (Arc(k,j))) or
             ((Arc(i,j)) and (Arc(i,k))));
  end; (* TwoOut *)
function ZeroIn: Boolean;
(* At least one in-degree 0. *)
  begin
  with m do
  ZeroIn :=(((not Arc(i,j)) and (not Arc(k,j))) or
            ((not Arc(i,k)) and (not Arc(j,k))) or
            ((not Arc(j,i)) and (not Arc(k,i))));
  end; (* ZeroIn *)

begin (* TriadCount *)
(* cc is a version of c with indices that make it more easily accessible *)
 for i:= 0 to 3 do for j:= 0 to 3 do for k:= 0 to 3 do cc[i,j,k] := 0;
with m do
 for i:= 1 to g do for j:= i+1 to g do for k:= j+1 to g do begin
 (* determine mut = number of mutuals in this triad;
              sum = number of arcs in this triad *)
   mut := 0; sum := 0;
   if (Arc(i,j)) then begin
    Inc(sum);
    if (Arc(j,i)) then
      begin Inc(mut); Inc(sum); end;
   end else if (Arc(j,i)) then Inc(sum);
  if (Arc(i,k)) then begin
    Inc(sum);
    if (Arc(k,i)) then
      begin Inc(mut); Inc(sum); end;
   end else if (Arc(k,i)) then Inc(sum);
  if (Arc(j,k)) then begin
    Inc(sum);
    if (Arc(k,j)) then
      begin Inc(mut); Inc(sum); end;
   end else if (Arc(k,j)) then Inc(sum);
  t[1] := mut;
  t[2] := sum - 2*mut  ;
  t[3] := 0;
  (* Because of the last assignment, only the triads MANX with X = T, U, or D
     need to have a further t[3] assignment. *)
   if (mut <= 0) then begin (*A1*)
    if (sum >= 3) then
      (* triadtype MAN = 030 *)
       begin if not Cycle then t[3] := 1; end
      else begin (*B*)
       if (sum >= 2) then begin (*C*)
         (* triadtype MAN = 021 *)
          if TwoIn then t[3] := 1 else
            begin if TwoOut then t[3] := 2; end;
          end (*C*) ; end (*B*) ;
      end (*A1*) else begin (*A2*)
        if (mut <= 1) then begin (*B*)
          if (sum >= 4) then begin (*C1*)
            (* triadtype MAN = 120 *)
           if not Cycle then
            begin if  ZeroIn then t[3] := 2 else t[3] := 1; end;
           end (*C1*) else begin (*C2*)
             if (sum >= 3 ) then
              (* triadtype MAN = 111 *)
              begin if TwoOut then t[3] := 1 else t[3] := 2; end;
             end (*C2*); end (*B*); end (*A2*);
        Inc(cc[t[1],t[2],t[3]]);
   end; (* of i,j,k loops; with m *)
  for i:=1 to 16 do c[i] := cc[TriadKind[i][1],TriadKind[i][2],TriadKind[i][3]];
end; (* TriadCount *)

procedure TriadProportion(var m: Adjac; g: Byte; var p: TriadVector);
var c: TriadCountType;
    i: Byte;
    sum: integer;
begin
  TriadCount(m,g,c);
  sum := g;
  sum := sum*(sum-1)*(sum-2) div 6;
  for i:=1 to 16 do p[i] := c[i]/sum;
end;

begin
InitTriad(TriadKind);
end. (* unit Triad *)